@@ -70,6 +70,7 @@
 //! to e.g. hotplug), a new Topology object should be created.
 
 use crate::misc::read_file_usize;
+use crate::misc::read_file_usize_vec;
 use crate::Cpumask;
 use anyhow::bail;
 use anyhow::Result;
@@ -115,6 +116,10 @@ pub struct Cpu {
     /// Base operational frqeuency. Only available on Intel Turbo Boost
     /// CPUs. If not available, this will simply return maximum frequency.
     pub base_freq: usize,
+    /// The best-effort guessting of cpu_capacity scaled to 1024
+    pub cpu_capacity: usize,
+    /// CPU idle resume latency
+    pub pm_qos_resume_latency_us: usize,
     pub trans_lat_ns: usize,
     pub l2_id: usize,
     pub l3_id: usize,
@@ -165,6 +170,7 @@ pub struct Llc {
 #[derive(Debug, Clone)]
 pub struct Node {
     pub id: usize,
+    pub distance: Vec<usize>,
     pub llcs: BTreeMap<usize, Arc<Llc>>,
     /// Cpumask of all CPUs in this node.
     pub span: Cpumask,
@@ -182,6 +188,8 @@ pub struct Topology {
     pub nodes: BTreeMap<usize, Node>,
     /// Cpumask all CPUs in the system.
     pub span: Cpumask,
+    /// True if SMT is enabled in the system, false otherwise.
+    pub smt_enabled: bool,
 
     /// Skip indices to access lower level members easily.
     pub all_llcs: BTreeMap<usize, Arc<Llc>>,
@@ -239,6 +247,7 @@ impl Topology {
         Ok(Topology {
             nodes,
             span,
+            smt_enabled: is_smt_active().unwrap_or(false),
             all_llcs: topo_llcs,
             all_cores: topo_cores,
             all_cpus: topo_cpus,
@@ -405,6 +414,7 @@ fn create_insert_cpu(
     topo_ctx: &mut TopoCtx,
     big_little: bool,
     avg_cpu_freq: Option<(usize, usize)>,
+    capacity_src: Option<(String, usize, usize)>,
     flatten_llc: bool,
 ) -> Result<()> {
     // CPU is offline. The Topology hierarchy is read-only, and assumes
@@ -447,6 +457,17 @@ fn create_insert_cpu(
     let base_freq = read_file_usize(&freq_path.join("base_frequency")).unwrap_or(max_freq);
     let trans_lat_ns = read_file_usize(&freq_path.join("cpuinfo_transition_latency")).unwrap_or(0);
 
+    // Cpu capacity
+    let (cap_suffix, _avg_rcap, max_rcap) = capacity_src.unwrap_or(("".to_string(), 1024, 1024));
+    let cap_path = cpu_path.join(cap_suffix);
+    let rcap = read_file_usize(&cap_path).unwrap_or(max_rcap);
+    let cpu_capacity = (rcap * 1024) / max_rcap;
+
+    // Power management
+    let power_path = cpu_path.join("power");
+    let pm_qos_resume_latency_us =
+        read_file_usize(&power_path.join("pm_qos_resume_latency_us")).unwrap_or(0);
+
     let num_llcs = topo_ctx.node_llc_kernel_ids.len();
     let llc_id = topo_ctx
         .node_llc_kernel_ids
@@ -507,6 +528,8 @@ fn create_insert_cpu(
             min_freq,
             max_freq,
             base_freq,
+            cpu_capacity,
+            pm_qos_resume_latency_us,
             trans_lat_ns,
             l2_id,
             l3_id,
@@ -548,6 +571,57 @@ fn read_cpu_ids() -> Result<Vec<usize>> {
     Ok(cpu_ids)
 }
 
+fn cpu_capacity_source() -> Option<(String, usize, usize)> {
+    // Sources for guessing cpu_capacity under /sys/devices/system/cpu/cpuX.
+    // They should be ordered from the most precise to the least precise.
+    let sources = [
+        "cpufreq/amd_pstate_highest_perf",
+        "acpi_cppc/highest_perf",
+        "cpu_capacity",
+        "cpufreq/cpuinfo_max_freq",
+    ];
+
+    // Find the most precise source for cpu_capacity estimation.
+    let prefix = "/sys/devices/system/cpu/cpu0";
+    let mut raw_capacity = 0;
+    let mut suffix = sources[sources.len() - 1];
+    for src in sources {
+        let path_str = [prefix, src].join("/");
+        let path = Path::new(&path_str);
+        raw_capacity = read_file_usize(&path).unwrap_or(0);
+        if raw_capacity > 0 {
+            suffix = src;
+            break;
+        }
+    }
+    if raw_capacity == 0 {
+        return None;
+    }
+
+    // Find the max raw_capacity value for scaling to 1024.
+    let mut max_raw_capacity = 0;
+    let mut avg_raw_capacity = 0;
+    let mut nr_cpus = 0;
+    let cpu_paths = glob("/sys/devices/system/cpu/cpu[0-9]*").ok()?;
+    for cpu_path in cpu_paths.filter_map(Result::ok) {
+        let raw_capacity = read_file_usize(&cpu_path.join(suffix)).unwrap_or(0);
+        if max_raw_capacity < raw_capacity {
+            max_raw_capacity = raw_capacity;
+        }
+        avg_raw_capacity += raw_capacity;
+        nr_cpus += 1;
+    }
+    if max_raw_capacity == 0 {
+        return None;
+    }
+
+    Some((
+        suffix.to_string(),
+        avg_raw_capacity / nr_cpus,
+        max_raw_capacity,
+    ))
+}
+
 // Return the average base frequency across all CPUs and the highest maximum frequency.
 fn avg_cpu_freq() -> Option<(usize, usize)> {
     let mut top_max_freq = 0;
@@ -585,6 +659,11 @@ fn has_big_little() -> Option<bool> {
     Some(clusters.len() > 1)
 }
 
+fn is_smt_active() -> Option<bool> {
+    let smt_on = read_file_usize(Path::new("/sys/devices/system/cpu/smt/active")).ok()?;
+    Some(smt_on == 1)
+}
+
 fn create_default_node(
     online_mask: &Cpumask,
     topo_ctx: &mut TopoCtx,
@@ -594,6 +673,7 @@ fn create_default_node(
 
     let mut node = Node {
         id: 0,
+        distance: vec![],
         llcs: BTreeMap::new(),
         span: Cpumask::new(),
         #[cfg(feature = "gpu-topology")]
@@ -616,6 +696,7 @@ fn create_default_node(
         bail!("/sys/devices/system/cpu sysfs node not found");
     }
 
+    let capacity_src = cpu_capacity_source();
     let avg_cpu_freq = avg_cpu_freq();
     let big_little = has_big_little().unwrap_or(false);
     let cpu_ids = read_cpu_ids()?;
@@ -627,6 +708,7 @@ fn create_default_node(
             topo_ctx,
             big_little,
             avg_cpu_freq,
+            capacity_src.clone(),
             flatten_llc,
         )?;
     }
@@ -654,9 +736,16 @@ fn create_numa_nodes(
                 bail!("Failed to parse NUMA node ID {}", numa_str);
             }
         };
-
+        let distance = read_file_usize_vec(
+            Path::new(&format!(
+                "/sys/devices/system/node/node{}/distance",
+                node_id
+            )),
+            ' ',
+        )?;
         let mut node = Node {
             id: node_id,
+            distance,
             llcs: BTreeMap::new(),
             span: Cpumask::new(),
 
@@ -679,6 +768,7 @@ fn create_numa_nodes(
         let cpu_pattern = numa_path.join("cpu[0-9]*");
         let cpu_paths = glob(cpu_pattern.to_string_lossy().as_ref())?;
         let big_little = has_big_little().unwrap_or(false);
+        let capacity_src = cpu_capacity_source();
         let avg_cpu_freq = avg_cpu_freq();
         for cpu_path in cpu_paths.filter_map(Result::ok) {
             let cpu_str = cpu_path.to_str().unwrap().trim();
@@ -696,6 +786,7 @@ fn create_numa_nodes(
                 topo_ctx,
                 big_little,
                 avg_cpu_freq,
+                capacity_src.clone(),
                 false,
             )?;
         }
