@@ -14,8 +14,6 @@ mod stats;
 use std::collections::BTreeMap;
 use std::ffi::c_int;
 use std::fmt::Write;
-use std::fs::File;
-use std::io::Read;
 use std::mem::MaybeUninit;
 use std::sync::atomic::AtomicBool;
 use std::sync::atomic::Ordering;
@@ -26,9 +24,6 @@ use anyhow::Context;
 use anyhow::Result;
 use clap::Parser;
 use crossbeam::channel::RecvTimeoutError;
-use libbpf_rs::skel::OpenSkel;
-use libbpf_rs::skel::Skel;
-use libbpf_rs::skel::SkelBuilder;
 use libbpf_rs::OpenObject;
 use libbpf_rs::ProgramInput;
 use log::warn;
@@ -37,8 +32,7 @@ use scx_stats::prelude::*;
 use scx_utils::autopower::{fetch_power_profile, PowerProfile};
 use scx_utils::build_id;
 use scx_utils::compat;
-use scx_utils::import_enums;
-use scx_utils::scx_enums;
+use scx_utils::pm::{cpu_idle_resume_latency_supported, update_cpu_idle_resume_latency};
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
@@ -141,6 +135,14 @@ struct Opts {
     #[clap(short = 'l', long, allow_hyphen_values = true, default_value = "20000")]
     slice_us_lag: i64,
 
+    /// Set CPU idle QoS resume latency in microseconds (-1 = disabled).
+    ///
+    /// Setting a lower latency value makes CPUs less likely to enter deeper idle states, enhancing
+    /// performance at the cost of higher power consumption. Alternatively, increasing the latency
+    /// value may reduce performance, but also improve power efficiency.
+    #[clap(short = 'I', long, allow_hyphen_values = true, default_value = "-1")]
+    idle_resume_us: i64,
+
     /// Disable preemption.
     ///
     /// Never allow tasks to preempt others before their assigned time slice expires. This can help
@@ -189,6 +191,14 @@ struct Opts {
     #[clap(long, action = clap::ArgAction::SetTrue)]
     disable_l3: bool,
 
+    /// Disable SMT awareness.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    disable_smt: bool,
+
+    /// Disable NUMA rebalancing.
+    #[clap(long, action = clap::ArgAction::SetTrue)]
+    disable_numa: bool,
+
     /// Enable CPU frequency control (only with schedutil governor).
     ///
     /// With this option enabled the CPU frequency will be automatically scaled based on the load.
@@ -228,20 +238,11 @@ struct Opts {
     help_stats: bool,
 }
 
-fn is_smt_active() -> std::io::Result<i32> {
-    let mut file = File::open("/sys/devices/system/cpu/smt/active")?;
-    let mut contents = String::new();
-    file.read_to_string(&mut contents)?;
-
-    let smt_active: i32 = contents.trim().parse().unwrap_or(0);
-
-    Ok(smt_active)
-}
-
 struct Scheduler<'a> {
     skel: BpfSkel<'a>,
     struct_ops: Option<libbpf_rs::Link>,
     opts: &'a Opts,
+    topo: Topology,
     power_profile: PowerProfile,
     stats_server: StatsServer<(), Metrics>,
     user_restart: bool,
@@ -254,11 +255,12 @@ impl<'a> Scheduler<'a> {
         // Validate command line arguments.
         assert!(opts.slice_us >= opts.slice_us_min);
 
+        // Initialize CPU topology.
+        let topo = Topology::new().unwrap();
+
         // Check host topology to determine if we need to enable SMT capabilities.
-        let smt_enabled = match is_smt_active() {
-            Ok(value) => value == 1,
-            Err(_) => false,
-        };
+        let smt_enabled = !opts.disable_smt && topo.smt_enabled;
+
         info!(
             "{} {} {}",
             SCHEDULER_NAME,
@@ -266,6 +268,20 @@ impl<'a> Scheduler<'a> {
             if smt_enabled { "SMT on" } else { "SMT off" }
         );
 
+        if opts.idle_resume_us >= 0 {
+            if !cpu_idle_resume_latency_supported() {
+                warn!("idle resume latency not supported");
+            } else {
+                info!("Setting idle QoS to {} us", opts.idle_resume_us);
+                for cpu in topo.all_cpus.values() {
+                    update_cpu_idle_resume_latency(
+                        cpu.id,
+                        opts.idle_resume_us.try_into().unwrap(),
+                    )?;
+                }
+            }
+        }
+
         // Initialize BPF connector.
         let mut skel_builder = BpfSkelBuilder::default();
         skel_builder.obj_builder.debug(opts.verbose);
@@ -276,6 +292,7 @@ impl<'a> Scheduler<'a> {
         // Override default BPF scheduling parameters.
         skel.maps.rodata_data.debug = opts.debug;
         skel.maps.rodata_data.smt_enabled = smt_enabled;
+        skel.maps.rodata_data.numa_disabled = opts.disable_numa;
         skel.maps.rodata_data.local_pcpu = opts.local_pcpu;
         skel.maps.rodata_data.local_kthreads = opts.local_kthreads;
         skel.maps.rodata_data.no_preempt = opts.no_preempt;
@@ -290,7 +307,8 @@ impl<'a> Scheduler<'a> {
         skel.struct_ops.bpfland_ops_mut().flags = *compat::SCX_OPS_ENQ_EXITING
             | *compat::SCX_OPS_ENQ_LAST
             | *compat::SCX_OPS_ENQ_MIGRATION_DISABLED
-            | *compat::SCX_OPS_BUILTIN_IDLE_PER_NODE;
+            | *compat::SCX_OPS_BUILTIN_IDLE_PER_NODE
+            | *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP;
         info!(
             "scheduler flags: {:#x}",
             skel.struct_ops.bpfland_ops_mut().flags
@@ -299,9 +317,6 @@ impl<'a> Scheduler<'a> {
         // Load the BPF program for validation.
         let mut skel = scx_ops_load!(skel, bpfland_ops, uei)?;
 
-        // Initialize CPU topology.
-        let topo = Topology::new().unwrap();
-
         // Initialize the primary scheduling domain and the preferred domain.
         let power_profile = fetch_power_profile(false);
         if let Err(err) = Self::init_energy_domain(&mut skel, &opts.primary_domain, power_profile) {
@@ -336,6 +351,7 @@ impl<'a> Scheduler<'a> {
             skel,
             struct_ops,
             opts,
+            topo,
             power_profile,
             stats_server,
             user_restart: false,
@@ -535,8 +551,8 @@ impl<'a> Scheduler<'a> {
 
         // Update the BPF cpumasks for the cache domains.
         for (cache_id, cpus) in cache_id_map {
-            // Ignore the cache domain if it includes a single CPU or all the CPUs.
-            if cpus.len() <= 1 || cpus.len() == *NR_CPU_IDS {
+            // Ignore the cache domain if it includes a single CPU.
+            if cpus.len() <= 1 {
                 continue;
             }
 
@@ -621,6 +637,16 @@ impl<'a> Scheduler<'a> {
 impl Drop for Scheduler<'_> {
     fn drop(&mut self) {
         info!("Unregister {} scheduler", SCHEDULER_NAME);
+
+        // Restore default CPU idle QoS resume latency.
+        if self.opts.idle_resume_us >= 0 {
+            if cpu_idle_resume_latency_supported() {
+                for cpu in self.topo.all_cpus.values() {
+                    update_cpu_idle_resume_latency(cpu.id, cpu.pm_qos_resume_latency_us as i32)
+                        .unwrap();
+                }
+            }
+        }
     }
 }
 
