@@ -112,6 +112,7 @@ struct task_ctx {
 	u64	wait_freq;		/* waiting frequency in a second */
 	u64	wake_freq;		/* waking-up frequency in a second */
 	u64	svc_time;		/* total CPU time consumed for this task */
+	u64	dsq_id;			/* DSQ id where a task run for statistics */
 
 	/*
 	 * Task deadline and time slice
@@ -120,7 +121,7 @@ struct task_ctx {
 	u32	lat_cri_waker;		/* waker's latency criticality */
 	u32	perf_cri;		/* performance criticality of a task */
 	u32	slice_ns;		/* time slice */
-	s8	futex_boost;		/* futex boost count */
+	s8	futex_boost;		/* futex acquired or not */
 	u8	is_greedy;		/* task's overscheduling ratio compared to its nice priority */
 	u8	need_lock_boost;	/* need to boost lock for deadline calculation */
 	u8	lock_holder_xted;	/* slice is already extended for a lock holder task */
@@ -128,6 +129,7 @@ struct task_ctx {
 	u8	slice_boost_prio;	/* how many times a task fully consumed the slice */
 	u8	on_big;			/* executable on a big core */
 	u8	on_little;		/* executable on a little core */
+	u8	is_affinitized;		/* is this task pinned to a subset of all CPUs? */
 };
 
 /*
