@@ -200,6 +200,8 @@ char _license[] SEC("license") = "GPL";
 #include "sys_stat.bpf.c"
 #include "preempt.bpf.c"
 #include "lock.bpf.c"
+#include "idle.bpf.c"
+#include "balance.bpf.c"
 
 /*
  * Logical current clock
@@ -286,6 +288,13 @@ static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
 	if (is_kernel_task(p))
 		weight_boost += LAVD_LC_WEIGHT_BOOST;
 
+	/*
+	 * Prioritize an affinitized task since it has restrictions
+	 * in placement so it tends to be delayed.
+	 */
+	if (taskc->is_affinitized)
+		weight_boost += LAVD_LC_WEIGHT_BOOST;
+
 	/*
 	 * Prioritize a migration-disabled task since it has restrictions
 	 * in placement so it tends to be delayed.
@@ -301,6 +310,9 @@ static u64 calc_weight_factor(struct task_struct *p, struct task_ctx *taskc)
 		weight_boost += LAVD_LC_WEIGHT_BOOST;
 	}
 
+	/*
+	 * Respect nice priority.
+	 */
 	weight_ft = p->scx.weight * weight_boost;
 	return weight_ft;
 }
@@ -383,7 +395,7 @@ static u64 calc_adj_runtime(u64 runtime)
 {
 	/*
 	 * Convert highly skewed runtime distribution to
-	 * mildlyskewed distribution.
+	 * mildly skewed distribution.
 	 */
 	u64 adj_runtime = log2_u64(runtime + 1);
 	return adj_runtime * adj_runtime;
@@ -425,6 +437,9 @@ static u64 calc_time_slice(struct task_struct *p, struct task_ctx *taskc)
 	u64 nr_queued;
 	u32 slice;
 
+	if (!taskc)
+		return LAVD_SLICE_MAX_NS_DFL;
+
 	/*
 	 * The time slice should be short enough to schedule all runnable tasks
 	 * at least once within a targeted latency.
@@ -498,11 +513,12 @@ static void advance_cur_logical_clk(struct task_struct *p)
 	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u64 vlc, clc, ret_clc;
 	u64 nr_queued, delta, new_clk;
+	int i;
 
 	vlc = READ_ONCE(p->scx.dsq_vtime);
 	clc = READ_ONCE(cur_logical_clk);
 
-	for (int i = 0; i < LAVD_MAX_RETRY; ++i) {
+	bpf_for(i, 0, LAVD_MAX_RETRY) {
 		/*
 		 * The clock should not go backward, so do nothing.
 		 */
@@ -535,6 +551,20 @@ static void update_stat_for_running(struct task_struct *p,
 	struct sys_stat *stat_cur = get_sys_stat_cur();
 	u64 wait_period, interval;
 
+	/*
+	 * If the sched_ext core directly dispatched a task, calculating the
+	 * task's deadline and time slice was also skipped. In this case, we
+	 * set the deadline and time slice here.
+	 *
+	 * Note that this is necessary when the kernel does not support
+	 * SCX_OPS_ENQ_MIGRATION_DISABLED or SCX_OPS_ENQ_MIGRATION_DISABLED
+	 * is not turned on.
+	 */
+	if (p->scx.slice == SCX_SLICE_DFL) {
+		p->scx.dsq_vtime = READ_ONCE(cur_logical_clk);
+		p->scx.slice = calc_time_slice(p, taskc);
+	}
+
 	/*
 	 * Update the current logical clock.
 	 */
@@ -592,6 +622,11 @@ static void update_stat_for_running(struct task_struct *p,
 	if (is_perf_cri(taskc, stat_cur))
 		cpuc->nr_perf_cri++;
 
+	if (taskc->dsq_id != cpuc->cpdom_id) {
+		taskc->dsq_id = cpuc->cpdom_id;
+		cpuc->nr_x_migration++;
+	}
+
 	/*
 	 * It is clear there is no need to consider the suspended duration
 	 * while running a task, so reset the suspended duration to zero.
@@ -623,6 +658,20 @@ static void update_stat_for_stopping(struct task_struct *p,
 
 	taskc->svc_time += task_run_time / p->scx.weight;
 
+	/*
+	 * Count how many times a task completely consumed the assigned time
+	 * slice to boost the task's slice. If not fully consumed, decrease
+	 * the slice boost priority by half.
+	 */
+	if (task_run_time > 0 && task_run_time >= taskc->slice_ns) {
+		if (taskc->slice_boost_prio < LAVD_SLICE_BOOST_MAX_STEP)
+			taskc->slice_boost_prio++;
+	}
+	else {
+		if (taskc->slice_boost_prio)
+			taskc->slice_boost_prio >>= 1;
+	}
+
 	/*
 	 * Reset waker's latency criticality here to limit the latency boost of
 	 * a task. A task will be latency-boosted only once after wake-up.
@@ -648,316 +697,6 @@ static void update_stat_for_stopping(struct task_struct *p,
 	taskc->lock_holder_xted = false;
 }
 
-static bool match_task_core_type(struct task_ctx *taskc,
-				 struct cpu_ctx *cpuc_prev,
-				 struct sys_stat *stat_cur)
-{
-	/*
-	 * If a task is performance critical, it is better to run on a big core
-	 * even paying some cost looking for a big core.
-	 */
-	if (is_perf_cri(taskc, stat_cur) && !cpuc_prev->big_core)
-		return false;
-
-	/*
-	 * Otherwise, it doesn't matter where it runs.
-	 */
-	return true;
-}
-
-static __always_inline
-bool could_run_on(struct task_struct *p, s32 cpu,
-			 struct bpf_cpumask *a_cpumask,
-			 struct bpf_cpumask *o_cpumask)
-{
-	bool ret;
-
-	ret = bpf_cpumask_test_cpu(cpu, p->cpus_ptr) &&
-	      (bpf_cpumask_test_cpu(cpu, cast_mask(a_cpumask)) ||
-	       bpf_cpumask_test_cpu(cpu, cast_mask(o_cpumask)));
-
-	return ret;
-}
-
-static __always_inline
-bool test_and_clear_cpu_idle(s32 cpu, const struct cpumask *idle_mask,
-			     bool reserve_cpu)
-{
-	if (reserve_cpu)
-		return scx_bpf_test_and_clear_cpu_idle(cpu);
-	return bpf_cpumask_test_cpu(cpu, idle_mask);
-}
-
-static __always_inline
-s32 find_idle_cpu_in(struct bpf_cpumask *cpumask,
-		     const struct cpumask *idle_mask, bool reserve_cpu)
-{
-	s32 cpu_id;
-
-	if (reserve_cpu) {
-		/*
-		 * Pick a fully idle core within a cpumask, then pick an
-		 * any idle core if there is no.
-		 */
-		cpu_id = scx_bpf_pick_idle_cpu(cast_mask(cpumask), 0);
-	} else {
-		cpu_id = bpf_cpumask_any_and_distribute(cast_mask(cpumask), idle_mask);
-		if (cpu_id >= nr_cpu_ids)
-			cpu_id = -EBUSY;
-	}
-
-	return cpu_id;
-}
-
-static __always_inline
-s32 find_idle_cpu(struct task_struct *p, struct task_ctx *taskc, s32 prev_cpu,
-		  u64 wake_flags, bool reserve_cpu, bool *is_idle)
-{
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-	struct cpu_ctx *cpuc, *cpuc_prev, *cpuc_waker;
-	struct bpf_cpumask *a_cpumask, *o_cpumask, *t_cpumask, *t2_cpumask;
-	struct bpf_cpumask *active, *ovrflw, *big, *little;
-	struct bpf_cpumask *cpdom_mask_prev, *cpdom_mask_waker;
-	const struct cpumask *idle_mask;
-	s32 cpu_id, waker_cpu;
-	int cpdom_id;
-
-	idle_mask = scx_bpf_get_idle_cpumask();
-
-	/*
-	 * If there is no idle cpu, stay on the previous cpu.
-	 */
-	if (!have_idle_cpus(idle_mask)) {
-		cpu_id = prev_cpu;
-		goto out;
-	}
-
-	/*
-	 * If a task can run only on a single CPU (e.g., per-CPU kworker), we
-	 * simply check if a task is still pinned on the prev_cpu and go.
-	 */
-	if (is_per_cpu_task(p) &&
-	    bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr)) {
-		if (test_and_clear_cpu_idle(prev_cpu, idle_mask, reserve_cpu))
-			*is_idle = true;
-		cpu_id = prev_cpu;
-		goto out;
-	}
-
-	/*
-	 * Prepare cpumaks.
-	 */
-	bpf_rcu_read_lock();
-
-	cpuc = get_cpu_ctx();
-	cpuc_prev = get_cpu_ctx_id(prev_cpu);
-	if (!cpuc || !cpuc_prev || !taskc) {
-		scx_bpf_error("Failed to lookup the current cpu_ctx");
-		cpu_id = -ENOENT;
-		goto unlock_out;
-	}
-
-	a_cpumask = cpuc->tmp_a_mask;
-	o_cpumask = cpuc->tmp_o_mask;
-	t_cpumask = cpuc->tmp_t_mask;
-	t2_cpumask = cpuc->tmp_t2_mask;
-	active  = active_cpumask;
-	ovrflw  = ovrflw_cpumask;
-	big = big_cpumask;
-	little = little_cpumask;
-	if (!a_cpumask || !o_cpumask || !t_cpumask || !t2_cpumask ||
-	    !active || !ovrflw || !big || !little) {
-		cpu_id = -ENOENT;
-		goto unlock_out;
-	}
-
-	cpdom_id = cpuc_prev->cpdom_id;
-	cpdom_mask_prev = MEMBER_VPTR(cpdom_cpumask, [cpdom_id]);
-	if (!cpdom_mask_prev) {
-		scx_bpf_error("Failed to lookup cpdom_cpumask for %d",
-			      cpuc_prev->cpdom_id);
-		cpu_id = -ENOENT;
-		goto unlock_out;
-	}
-
-	cpuc_waker = get_cpu_ctx();
-	if (!cpuc_waker) {
-		scx_bpf_error("Failed to lookup the current cpu_ctx");
-		cpu_id = -ENOENT;
-		goto unlock_out;
-	}
-	waker_cpu = cpuc_waker->cpu_id;
-
-	cpdom_id = cpuc_waker->cpdom_id;
-	cpdom_mask_waker = MEMBER_VPTR(cpdom_cpumask, [cpdom_id]);
-	if (!cpdom_mask_waker) {
-		scx_bpf_error("Failed to lookup cpdom_cpumask for %d",
-			      cpuc_waker->cpdom_id);
-		cpu_id = -ENOENT;
-		goto unlock_out;
-	}
-
-	bpf_cpumask_and(a_cpumask, p->cpus_ptr, cast_mask(active));
-	bpf_cpumask_and(o_cpumask, p->cpus_ptr, cast_mask(ovrflw));
-
-	/*
-	 * Try to stay on the previous core if it is on active or ovrfw.
-	 */
-	if (match_task_core_type(taskc, cpuc_prev, stat_cur) &&
-	    could_run_on(p, prev_cpu, a_cpumask, o_cpumask) &&
-	    test_and_clear_cpu_idle(prev_cpu, idle_mask, reserve_cpu)) {
-		cpu_id = prev_cpu;
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * Try to stay on the waker's core if it is on active or ovrfw.
-	 */
-	if (wake_flags & SCX_WAKE_SYNC && prev_cpu != waker_cpu &&
-	    match_task_core_type(taskc, cpuc_waker, stat_cur) &&
-	    could_run_on(p, waker_cpu, a_cpumask, o_cpumask) &&
-	    test_and_clear_cpu_idle(waker_cpu, idle_mask, reserve_cpu)) {
-		cpu_id = waker_cpu;
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * Find cpumasks for a matching core type and LLC domain.
-	 */
-	if (bpf_cpumask_empty(cast_mask(a_cpumask)))
-		goto start_omask;
-
-	if (!have_little_core || is_perf_cri(taskc, stat_cur) || no_core_compaction) {
-		bpf_cpumask_and(t_cpumask, cast_mask(a_cpumask), cast_mask(big));
-	}
-	else {
-		bpf_cpumask_and(t_cpumask, cast_mask(a_cpumask), cast_mask(little));
-		goto start_llc_mask;
-	}
-
-	/*
-	 * Pick an idle core among turbo boost-enabled CPUs with a matching
-	 * core type.
-	 */
-	if (!have_turbo_core || no_prefer_turbo_core || !turbo_cpumask)
-		goto start_llc_mask;
-
-	bpf_cpumask_and(t2_cpumask, cast_mask(t_cpumask), cast_mask(turbo_cpumask));
-	if (bpf_cpumask_empty(cast_mask(t2_cpumask)))
-		goto start_llc_mask;
-
-	cpu_id = find_idle_cpu_in(t2_cpumask, idle_mask, reserve_cpu);
-	if (cpu_id >= 0) {
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * Pick an idle core among active CPUs with a matching core type within
-	 * the prev CPU's LLC domain.
-	 */
-start_llc_mask:
-	bpf_cpumask_and(t2_cpumask, cast_mask(t_cpumask), cast_mask(cpdom_mask_prev));
-	if (bpf_cpumask_empty(cast_mask(t2_cpumask)))
-		goto start_tmask;
-
-	cpu_id = find_idle_cpu_in(t2_cpumask, idle_mask, reserve_cpu);
-	if (cpu_id >= 0) {
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * Pick an idle core among active CPUs with a matching core type within
-	 * the waker CPU's LLC domain.
-	 */
-	if (wake_flags & SCX_WAKE_SYNC && prev_cpu != waker_cpu) {
-		bpf_cpumask_and(t2_cpumask, cast_mask(t_cpumask), cast_mask(cpdom_mask_waker));
-		if (bpf_cpumask_empty(cast_mask(t2_cpumask)))
-			goto start_tmask;
-
-		cpu_id = find_idle_cpu_in(t2_cpumask, idle_mask, reserve_cpu);
-		if (cpu_id >= 0) {
-			*is_idle = true;
-			goto unlock_out;
-		}
-	}
-
-	/*
-	 * Pick an idle core among active CPUs with a matching core type.
-	 */
-start_tmask:
-	if (have_little_core) {
-		cpu_id = find_idle_cpu_in(t_cpumask, idle_mask, reserve_cpu);
-		if (cpu_id >= 0) {
-			*is_idle = true;
-			goto unlock_out;
-		}
-	}
-
-	/*
-	 * Pick a idle core among active CPUs.
-	 */
-	cpu_id = find_idle_cpu_in(a_cpumask, idle_mask, reserve_cpu);
-	if (cpu_id >= 0) {
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * Pick an any idle core among overflow CPUs.
-	 */
-start_omask:
-	if (bpf_cpumask_empty(cast_mask(o_cpumask)))
-		goto start_any_mask;
-
-	cpu_id = find_idle_cpu_in(o_cpumask, idle_mask, reserve_cpu);
-	if (cpu_id >= 0) {
-		*is_idle = true;
-		goto unlock_out;
-	}
-
-	/*
-	 * If there is no idle core under our control, pick random core
-	 * either in active of overflow CPUs.
-	 */
-	if (!bpf_cpumask_empty(cast_mask(a_cpumask))) {
-		cpu_id = bpf_cpumask_any_distribute(cast_mask(a_cpumask));
-		goto unlock_out;
-	}
-
-	if (!bpf_cpumask_empty(cast_mask(o_cpumask))) {
-		cpu_id = bpf_cpumask_any_distribute(cast_mask(o_cpumask));
-		goto unlock_out;
-	}
-
-	/*
-	 * If the task cannot run on either active or overflow cores,
-	 * stay on the previous core (if it is okay) or one of its taskset.
-	 * Then, put the CPU to the overflow set.
-	 */
-start_any_mask:
-	if (bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr))
-		cpu_id = prev_cpu;
-	else
-		cpu_id = bpf_cpumask_any_distribute(p->cpus_ptr);
-	bpf_cpumask_set_cpu(cpu_id, ovrflw);
-
-unlock_out:
-	bpf_rcu_read_unlock();
-
-out:
-	scx_bpf_put_idle_cpumask(idle_mask);
-
-	/*
-	 * Note that we don't need to kick the picked CPU here since the
-	 * ops.select_cpu() path internally triggers kicking cpu if necessary.
-	 */
-	return cpu_id;
-}
-
 static u64 calc_when_to_run(struct task_struct *p, struct task_ctx *taskc)
 {
 	u64 deadline_delta;
@@ -974,12 +713,17 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 		   u64 wake_flags)
 {
 	bool found_idle = false;
-	struct task_ctx *taskc;
+	struct task_ctx *taskc = get_task_ctx(p);
 	struct cpu_ctx *cpuc;
 	u64 dsq_id;
 	s32 cpu_id;
+	struct pick_ctx ictx = {
+		.p = p,
+		.taskc = taskc,
+		.prev_cpu = prev_cpu,
+		.wake_flags = wake_flags,
+	};
 
-	taskc = get_task_ctx(p);
 	if (!taskc)
 		return prev_cpu;
 	taskc->wakeup_ft += !!(wake_flags & SCX_WAKE_SYNC);
@@ -988,7 +732,7 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 	 * Find an idle cpu and reserve it since the task @p will run
 	 * on the idle cpu.
 	 */
-	cpu_id = find_idle_cpu(p, taskc, prev_cpu, wake_flags, true, &found_idle);
+	cpu_id = pick_idle_cpu(&ictx, &found_idle);
 	if (found_idle) {
 		/*
 		 * If there is an idle cpu and its associated DSQ is empty,
@@ -1015,67 +759,27 @@ s32 BPF_STRUCT_OPS(lavd_select_cpu, struct task_struct *p, s32 prev_cpu,
 	return cpu_id >= 0 ? cpu_id : prev_cpu;
 }
 
-static u64 find_proper_dsq(struct task_ctx *taskc, struct cpu_ctx *cpuc)
+static bool can_direct_dispatch(u64 dsq_id, s32 cpu, bool is_idle)
 {
-	struct sys_stat *stat_cur = get_sys_stat_cur();
-	bool perf_cri = is_perf_cri(taskc, stat_cur);
-	bool big_core = cpuc->big_core;
-
 	/*
-	 * If a task type and a core type matches, use the current cpu's
-	 * compute domain DSQ.
+	 * If the chosen CPU is idle and there is nothing to do
+	 * in the domain, we can safely choose the fast track.
 	 */
-	if (perf_cri == big_core)
-		return cpuc->cpdom_id;
-
-	/*
-	 * Otherwise, use the DSQ of an alternative core type.
-	 */
-	return cpuc->cpdom_alt_id;
-}
-
-static bool try_kick_task_idle_cpu(struct task_struct *p,
-				   struct task_ctx *taskc, s32 prev_cpu)
-{
-	bool found_idle = false;
-	s32 cpu;
-
-	/*
-	 * Find an idle cpu but do not reserve the idle cpu. That is because
-	 * there is no guarantee the idle cpu will be picked up at this point.
-	 */
-	cpu = find_idle_cpu(p, taskc, prev_cpu, 0, false, &found_idle);
-	if (found_idle && cpu >= 0) {
-		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+	if (is_idle && cpu >= 0 && !scx_bpf_dsq_nr_queued(dsq_id))
 		return true;
-	}
-
 	return false;
 }
 
 void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 {
-	struct cpu_ctx *cpuc_task, *cpuc_cur;
+	struct cpu_ctx *cpuc_cur;
 	struct task_ctx *taskc;
-	s32 prev_cpu;
+	s32 task_cpu, cpu = -ENOENT;
 	u64 dsq_id, now;
+	bool is_idle = false;
 
-	/*
-	 * Place a task to a run queue of current cpu's compute domain.
-	 *
-	 * If there is an idle CPU at the ops.select_cpu(), the task is already
-	 * dispatched at ops.select_cpu(), so ops.enqueue() won't be called.
-	 * Hence, the task that is enqueued here are the cases: 1) there is no
-	 * idle CPU when ops.select_cpu() or 2) the task is not the case of
-	 * being wakened up (i.e., resume after preemption). Therefore, we
-	 * always put the task to the global DSQ, so any idle CPU can pick it
-	 * up.
-	 */
 	taskc = get_task_ctx(p);
-	prev_cpu = scx_bpf_task_cpu(p);
-	cpuc_task = get_cpu_ctx_id(prev_cpu);
-	cpuc_cur = get_cpu_ctx();
-	if (!cpuc_cur || !cpuc_task || !taskc)
+	if (!taskc)
 		return;
 
 	/*
@@ -1086,222 +790,45 @@ void BPF_STRUCT_OPS(lavd_enqueue, struct task_struct *p, u64 enq_flags)
 	p->scx.slice = calc_time_slice(p, taskc);
 
 	/*
-	 * Enqueue the task to one of task's DSQs based on its virtual deadline.
-	 *
-	 * We do not perform direct dispatch to a local DSQ (SCX_DSQ_LOCAL_ON)
-	 * on purpose. In particular, the direct dispatch at SCX_ENQ_CPU_SELECTED
-	 * could increase tail latencies because it gives too much favor to
-	 * non-migratable tasks, so stalling others.
+	 * Find a proper DSQ for the task, which is either the task's
+	 * associated compute domain or its alternative domain, or
+	 * the closest available domain from the previous domain.
 	 */
-	dsq_id = find_proper_dsq(taskc, cpuc_task);
-	prev_cpu = scx_bpf_task_cpu(p);
-	scx_bpf_dsq_insert_vtime(p, dsq_id, p->scx.slice, p->scx.dsq_vtime,
-				 enq_flags);
+	task_cpu = scx_bpf_task_cpu(p);
+	dsq_id = pick_proper_dsq(p, taskc, task_cpu, &cpu, &is_idle);
 
 	/*
-	 * If there is an idle cpu for the task, try to kick it up now
-	 * so it can consume the task immediately.
+	 * Enqueue the task to a DSQ. If it is safe to directly dispatch
+	 * to the local DSQ of the chosen CPU, do it. Otherwise, enqueue
+	 * to the chosen DSQ of the chosen domain.
 	 */
-	if (try_kick_task_idle_cpu(p, taskc, prev_cpu))
+	if (can_direct_dispatch(dsq_id, cpu, is_idle)) {
+		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL_ON | cpu, p->scx.slice,
+				   enq_flags);
+	} else {
+		scx_bpf_dsq_insert_vtime(p, dsq_id, p->scx.slice,
+					 p->scx.dsq_vtime, enq_flags);
+	}
+
+	/*
+	 * If a new overflow CPU was assigned while finding a proper DSQ,
+	 * kick the new CPU and go.
+	 */
+	if (is_idle && cpu >= 0) {
+		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 		return;
-
-	/*
-	 * If there is no idle cpu for an eligible task, try to preempt a task.
-	 * Try to find and kick a victim CPU, which runs a less urgent task.
-	 * The kick will be done asynchronously.
-	 */
-	now = scx_bpf_now();
-	try_find_and_kick_victim_cpu(p, taskc, cpuc_cur, dsq_id, now);
-}
-
-static bool consume_dsq(u64 dsq_id)
-{
-	struct cpdom_ctx *cpdomc;
-
-	/*
-	 * Update the last consume clock of the compute domain.
-	 */
-	cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-	if (!cpdomc) {
-		scx_bpf_error("Failed to lookup cpdom_ctx for %llu", dsq_id);
-		return false;
 	}
 
 	/*
-	 * Try to consume a task on the associated DSQ.
+	 * If there is no idle CPU for an eligible task, try to preempt a task.
+	 * Try to find and kick a victim CPU, which runs a less urgent task,
+	 * from dsq_id. The kick will be done asynchronously.
 	 */
-	return scx_bpf_dsq_move_to_local(dsq_id);
-}
-
-static bool try_to_steal_task(struct cpdom_ctx *cpdomc)
-{
-	struct cpdom_ctx *cpdomc_pick;
-	u64 nr_nbr, dsq_id;
-	s64 nuance;
-
-	/*
-	 * If all CPUs are not used -- i.e., the system is under-utilized,
-	 * there is no point of load balancing. It is better to make an
-	 * effort to increase the system utilization.
-	 */
-	if (!use_full_cpus())
-		return false;
-
-	/*
-	 * Probabilistically make a go or no go decision to avoid the
-	 * thundering herd problem. In other words, one out of nr_cpus
-	 * will try to steal a task at a moment.
-	 */
-	if (!prob_x_out_of_y(1, cpdomc->nr_cpus * LAVD_CPDOM_X_PROB_FT))
-		return false;
-
-	/*
-	 * Traverse neighbor compute domains in distance order.
-	 */
-	nuance = bpf_get_prandom_u32();
-	for (int i = 0; i < LAVD_CPDOM_MAX_DIST; i++) {
-		nr_nbr = min(cpdomc->nr_neighbors[i], LAVD_CPDOM_MAX_NR);
-		if (nr_nbr == 0)
-			break;
-
-		/*
-		 * Traverse neighbor in the same distance in arbitrary order.
-		 */
-		for (int j = 0; j < LAVD_CPDOM_MAX_NR; j++, nuance = dsq_id + 1) {
-			if (j >= nr_nbr)
-				break;
-
-			dsq_id = pick_any_bit(cpdomc->neighbor_bits[i], nuance);
-			if (dsq_id == -ENOENT)
-				continue;
-
-			cpdomc_pick = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-			if (!cpdomc_pick) {
-				scx_bpf_error("Failed to lookup cpdom_ctx for %llu", dsq_id);
-				return false;
-			}
-
-			if (!cpdomc_pick->is_stealee || !cpdomc_pick->is_active)
-				continue;
-
-			/*
-			 * If task stealing is successful, mark the stealer
-			 * and the stealee's job done. By marking done,
-			 * those compute domains would not be involved in
-			 * load balancing until the end of this round,
-			 * so this helps gradual migration. Note that multiple
-			 * stealers can steal tasks from the same stealee.
-			 * However, we don't coordinate concurrent stealing
-			 * because the chance is low and there is no harm
-			 * in slight over-stealing.
-			 */
-			if (consume_dsq(dsq_id)) {
-				WRITE_ONCE(cpdomc_pick->is_stealee, false);
-				WRITE_ONCE(cpdomc->is_stealer, false);
-				return true;
-			}
-		}
-
-		/*
-		 * Now, we need to steal a task from a farther neighbor
-		 * for load balancing. Since task migration from a farther
-		 * neighbor is more expensive (e.g., crossing a NUMA boundary),
-		 * we will do this with a lot of hesitation. The chance of
-		 * further migration will decrease exponentially as distance
-		 * increases, so, on the other hand, it increases the chance
-		 * of closer migration.
-		 */
-		if (!prob_x_out_of_y(1, LAVD_CPDOM_X_PROB_FT))
-			break;
+	cpuc_cur = get_cpu_ctx();
+	if (cpuc_cur) {
+		now = scx_bpf_now();
+		try_find_and_kick_victim_cpu(p, taskc, cpuc_cur, dsq_id, now);
 	}
-
-	return false;
-}
-
-static bool force_to_steal_task(struct cpdom_ctx *cpdomc)
-{
-	struct cpdom_ctx *cpdomc_pick;
-	u64 nr_nbr, dsq_id;
-	s64 nuance;
-
-	/*
-	 * Traverse neighbor compute domains in distance order.
-	 */
-	nuance = bpf_get_prandom_u32();
-	for (int i = 0; i < LAVD_CPDOM_MAX_DIST; i++) {
-		nr_nbr = min(cpdomc->nr_neighbors[i], LAVD_CPDOM_MAX_NR);
-		if (nr_nbr == 0)
-			break;
-
-		/*
-		 * Traverse neighbor in the same distance in arbitrary order.
-		 */
-		for (int j = 0; j < LAVD_CPDOM_MAX_NR; j++, nuance = dsq_id + 1) {
-			if (j >= nr_nbr)
-				break;
-
-			dsq_id = pick_any_bit(cpdomc->neighbor_bits[i], nuance);
-			if (dsq_id == -ENOENT)
-				continue;
-
-			cpdomc_pick = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-			if (!cpdomc_pick) {
-				scx_bpf_error("Failed to lookup cpdom_ctx for %llu", dsq_id);
-				return false;
-			}
-
-			if (!cpdomc_pick->is_active)
-				continue;
-
-			if (consume_dsq(dsq_id))
-				return true;
-		}
-	}
-
-	return false;
-}
-
-static bool consume_task(struct cpu_ctx *cpuc)
-{
-	struct cpdom_ctx *cpdomc;
-	u64 dsq_id;
-
-	dsq_id = cpuc->cpdom_id;
-	cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-	if (!cpdomc) {
-		scx_bpf_error("Failed to lookup cpdom_ctx for %llu", dsq_id);
-		return false;
-	}
-
-	/*
-	 * If the current compute domain is a stealer, try to steal
-	 * a task from any of stealee domains probabilistically.
-	 */
-	if (READ_ONCE(cpdomc->is_stealer) && try_to_steal_task(cpdomc))
-		goto x_domain_migration_out;
-
-	/*
-	 * Try to consume a task from CPU's associated DSQ.
-	 */
-	if (consume_dsq(dsq_id))
-		return true;
-
-	/*
-	 * If there is no task in the assssociated DSQ, traverse neighbor
-	 * compute domains in distance order -- task stealing.
-	 */
-	if (force_to_steal_task(cpdomc))
-		goto x_domain_migration_out;
-
-	return false;
-
-	/*
-	 * Task migration across compute domains happens.
-	 * Update the statistics.
-	 */
-x_domain_migration_out:
-	cpuc->nr_x_migration++;
-	return true;
 }
 
 static bool try_continue_lock_holder(struct task_struct *prev,
@@ -1330,20 +857,21 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 {
 	struct cpu_ctx *cpuc;
 	struct task_ctx *taskc;
-	struct bpf_cpumask *active, *ovrflw;
+	struct bpf_cpumask *active, *ovrflw, *cpdom_mask, *new_mask;
 	struct task_struct *p;
 	u64 dsq_id;
+	s32 new_cpu;
 	bool try_consume = false;
 
 	cpuc = get_cpu_ctx_id(cpu);
 	if (!cpuc) {
-		scx_bpf_error("Failed to look up cpu context context");
+		scx_bpf_error("Failed to look up cpu context");
 		return;
 	}
 	dsq_id = cpuc->cpdom_id;
 
 	/*
-	 * If a task newly holds a lock, continue to execute it
+	 * If a task is holding a new lock, continue to execute it
 	 * to make system-wide forward progress.
 	 */
 	if (prev) {
@@ -1352,7 +880,6 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 			scx_bpf_error("Failed to look up task context");
 			return;
 		}
-
 		if (try_continue_lock_holder(prev, taskc, cpuc))
 			return;
 	}
@@ -1372,97 +899,130 @@ void BPF_STRUCT_OPS(lavd_dispatch, s32 cpu, struct task_struct *prev)
 
 	active = active_cpumask;
 	ovrflw = ovrflw_cpumask;
-	if (!active || !ovrflw) {
+	cpdom_mask = MEMBER_VPTR(cpdom_cpumask, [cpuc->cpdom_id]);
+	new_mask = cpuc->tmp_t_mask;
+	if (!active || !ovrflw || !cpdom_mask || !new_mask) {
 		scx_bpf_error("Failed to prepare cpumasks.");
 		goto unlock_out;
 	}
 
 	/*
-	 * If the CPU belonges to the active or overflow set, dispatch a task.
+	 * If the current CPU belonges to either active or overflow set,
+	 * dispatch a task and go.
 	 */
 	if (bpf_cpumask_test_cpu(cpu, cast_mask(active)) ||
 	    bpf_cpumask_test_cpu(cpu, cast_mask(ovrflw))) {
 		try_consume = true;
 		goto unlock_out;
 	}
+	/* NOTE: This CPU belongs to neither active nor overflow set. */
 
 	/*
-	 * If this CPU is not either in active or overflow CPUs, it tries to
-	 * find and run a task pinned to run on this CPU.
+	 * If the previous task is a per-CPU task on this CPU,
+	 * extend the overflow set and go.
+	 */
+	if (prev && is_per_cpu_task(prev)) {
+		bpf_cpumask_set_cpu(cpu, ovrflw);
+		try_consume = true;
+		goto unlock_out;
+	}
+
+	/*
+	 * If the previous task can run on this CPU but not on either active
+	 * or overflow set, extend the overflow set and go.
+	 */
+	if (prev && taskc->is_affinitized &&
+	    bpf_cpumask_test_cpu(cpu, prev->cpus_ptr) &&
+	    !bpf_cpumask_intersects(cast_mask(active), prev->cpus_ptr) &&
+	    !bpf_cpumask_intersects(cast_mask(ovrflw), prev->cpus_ptr)) {
+		bpf_cpumask_set_cpu(cpu, ovrflw);
+		try_consume = true;
+		goto unlock_out;
+	}
+
+	/*
+	 * If this CPU is neither in active nor overflow CPUs,
+	 * try to find and run the task affinitized on this CPU.
 	 */
 	bpf_for_each(scx_dsq, p, dsq_id, 0) {
 		/*
-		 * Prioritize kernel tasks because most kernel tasks are pinned
-		 * to a particular CPU and latency-critical (e.g., ksoftirqd,
-		 * kworker, etc).
-		 */
-		if (is_kernel_task(p)) {
-			bpf_cpumask_set_cpu(cpu, ovrflw);
-			try_consume = true;
-			break;
-		}
-
-		/*
-		 * This is a hack to bypass the restriction of the current BPF
-		 * not trusting the pointer p. Once the BPF verifier gets
-		 * smarter, we can remove bpf_task_from_pid().
+		 * Note that this is a hack to bypass the restriction of the
+		 * current BPF not trusting the pointer p. Once the BPF
+		 * verifier gets smarter, we can remove bpf_task_from_pid().
 		 */
 		p = bpf_task_from_pid(p->pid);
 		if (!p)
-			goto unlock_out;
+			break;
 
 		/*
-		 * If a task can run on active or overflow CPUs, it just does
-		 * nothing to go idle.
+		 * If the task is a per-CPU task on this CPU,
+		 * extend the overflow set and go.
+		 * But not on this CPU, try another task.
 		 */
-		if (bpf_cpumask_intersects(cast_mask(active), p->cpus_ptr) ||
-		    bpf_cpumask_intersects(cast_mask(ovrflw), p->cpus_ptr))
-			goto release_break;
+		if (is_per_cpu_task(p)) {
+			new_cpu = scx_bpf_task_cpu(p);
+			if (new_cpu == cpu) {
+				bpf_cpumask_set_cpu(new_cpu, ovrflw);
+				bpf_task_release(p);
+				try_consume = true;
+				break;
+			}
+			else if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
+				scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
+
+			bpf_task_release(p);
+			continue;
+		}
 
 		/*
-		 * This is the first time a particular pinned user-space task
-		 * is run on this CPU at this interval. From now on, this CPU
-		 * will be part of the overflow CPU so can be used to run the
-		 * pinned task and the other tasks. Note that we don't need to
-		 * kick @cpu here since @cpu is the current CPU, which is
-		 * obviously not idle.
+		 * If the task can run on either active or overflow set,
+		 * try another task.
 		 */
-		bpf_cpumask_set_cpu(cpu, ovrflw);
+		taskc = get_task_ctx(p);
+		if(taskc && (!taskc->is_affinitized ||
+		   bpf_cpumask_intersects(cast_mask(active), p->cpus_ptr) ||
+		   bpf_cpumask_intersects(cast_mask(ovrflw), p->cpus_ptr))) {
+			bpf_task_release(p);
+			continue;
+		}
 
 		/*
-		 * Otherwise, that means there is a task that should run on
-		 * this particular CPU. So, consume one of such tasks.
-		 *
-		 * Note that this path is not optimized since
-		 * scx_bpf_dsq_move_to_local() should traverse until it finds
-		 * any task that can run on this CPU. The scheduled task might
-		 * be runnable on the active cores. We will optimize this path
-		 * after introducing per-core DSQ.
+		 * Now, we know that the task cannot run on either active
+		 * or overflow set. Then, let's consider to extend the
+		 * overflow set.
 		 */
-		try_consume = true;
-
-release_break:
+		new_cpu = find_cpu_in(p->cpus_ptr, cpuc);
+		if (new_cpu >= 0) {
+			if (new_cpu == cpu) {
+				bpf_cpumask_set_cpu(new_cpu, ovrflw);
+				bpf_task_release(p);
+				try_consume = true;
+				break;
+			}
+			else if (!bpf_cpumask_test_and_set_cpu(new_cpu, ovrflw))
+				scx_bpf_kick_cpu(new_cpu, SCX_KICK_IDLE);
+		}
 		bpf_task_release(p);
-		break;
 	}
 
 unlock_out:
 	bpf_rcu_read_unlock();
 
-	/*
-	 * Note that the verifier in 6.8 kernel cannot correctly verifies the
-	 * code when consume_task() is under a rcu-read-lock region. Hence, we
-	 * moveed it outside of the rcu region. :-(
-	 */
 consume_out:
 	/*
-	 * Consume a task if requested.
+	 * If this CPU should go idle, do nothing.
 	 */
-	if (try_consume && consume_task(cpuc))
+	if (!try_consume)
 		return;
 
 	/*
-	 * If nothing to run, continue to run the previous task.
+	 * Otherwise, consume a task.
+	 */
+	if (consume_task(cpuc))
+		return;
+
+	/*
+	 * If nothing to run, continue running the previous task.
 	 */
 	if (prev && prev->scx.flags & SCX_TASK_QUEUED)
 		prev->scx.slice = calc_time_slice(prev, taskc);
@@ -1526,7 +1086,7 @@ void BPF_STRUCT_OPS(lavd_runnable, struct task_struct *p, u64 enq_flags)
 	 * Filter out unrelated tasks.
 	 */
 	waker = bpf_get_current_task_btf();
-	waker_taskc = try_get_task_ctx(waker);
+	waker_taskc = get_task_ctx(waker);
 	if (!waker_taskc) {
 		/*
 		 * In this case, the waker could be an idle task
@@ -1593,31 +1153,6 @@ void BPF_STRUCT_OPS(lavd_running, struct task_struct *p)
 	try_proc_introspec_cmd(p, taskc);
 }
 
-static bool slice_fully_consumed(struct cpu_ctx *cpuc, struct task_ctx *taskc)
-{
-	u64 run_time_ns;
-
-	run_time_ns = time_delta(taskc->last_stopping_clk, taskc->last_running_clk);
-	return run_time_ns > 0 && run_time_ns >= taskc->slice_ns;
-}
-
-static void adjust_slice_boost(struct cpu_ctx *cpuc, struct task_ctx *taskc)
-{
-	/*
-	 * Count how many times a task completely consumed the assigned time
-	 * slice to boost the task's slice when CPU is under-utilized. If not
-	 * fully consumed, decrease the slice boost priority by half.
-	 */
-	if (slice_fully_consumed(cpuc, taskc)) {
-		if (taskc->slice_boost_prio < LAVD_SLICE_BOOST_MAX_STEP)
-			taskc->slice_boost_prio++;
-	}
-	else {
-		if (taskc->slice_boost_prio)
-			taskc->slice_boost_prio >>= 1;
-	}
-}
-
 void BPF_STRUCT_OPS(lavd_stopping, struct task_struct *p, bool runnable)
 {
 	struct cpu_ctx *cpuc;
@@ -1632,11 +1167,6 @@ void BPF_STRUCT_OPS(lavd_stopping, struct task_struct *p, bool runnable)
 		return;
 
 	update_stat_for_stopping(p, taskc, cpuc);
-
-	/*
-	 * Adjust slice boost for the task's next schedule.
-	 */
-	adjust_slice_boost(cpuc, taskc);
 }
 
 void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
@@ -1672,8 +1202,17 @@ void BPF_STRUCT_OPS(lavd_quiescent, struct task_struct *p, u64 deq_flags)
 
 static void cpu_ctx_init_online(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
 {
+	struct bpf_cpumask *cd_cpumask;
+
+	bpf_rcu_read_lock();
+	cd_cpumask = MEMBER_VPTR(cpdom_cpumask, [cpuc->cpdom_id]);
+	if (!cd_cpumask)
+		goto unlock_out;
+	bpf_cpumask_set_cpu(cpu_id, cd_cpumask);
+unlock_out:
+	bpf_rcu_read_unlock();
+
 	cpuc->idle_start_clk = 0;
-	cpuc->cpu_id = cpu_id;
 	cpuc->lat_cri = 0;
 	cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
 	WRITE_ONCE(cpuc->online_clk, now);
@@ -1684,8 +1223,17 @@ static void cpu_ctx_init_online(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
 
 static void cpu_ctx_init_offline(struct cpu_ctx *cpuc, u32 cpu_id, u64 now)
 {
+	struct bpf_cpumask *cd_cpumask;
+
+	bpf_rcu_read_lock();
+	cd_cpumask = MEMBER_VPTR(cpdom_cpumask, [cpuc->cpdom_id]);
+	if (!cd_cpumask)
+		goto unlock_out;
+	bpf_cpumask_clear_cpu(cpu_id, cd_cpumask);
+unlock_out:
+	bpf_rcu_read_unlock();
+
 	cpuc->idle_start_clk = 0;
-	cpuc->cpu_id = cpu_id;
 	WRITE_ONCE(cpuc->offline_clk, now);
 	cpuc->is_online = false;
 	barrier();
@@ -1798,6 +1346,7 @@ void BPF_STRUCT_OPS(lavd_set_cpumask, struct task_struct *p,
 		return;
 	}
 
+	taskc->is_affinitized = bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids;
 	set_on_core_type(taskc, cpumask);
 }
 
@@ -1846,6 +1395,7 @@ static void init_task_ctx(struct task_struct *p, struct task_ctx *taskc)
 	u64 now = scx_bpf_now();
 
 	__builtin_memset(taskc, 0, sizeof(*taskc));
+	taskc->is_affinitized = bpf_cpumask_weight(p->cpus_ptr) != nr_cpu_ids;
 	taskc->last_running_clk = now; /* for run_time_ns */
 	taskc->last_stopping_clk = now; /* for run_time_ns */
 	taskc->run_time_ns = slice_max_ns;
@@ -1893,7 +1443,7 @@ static s32 init_cpdoms(u64 now)
 			scx_bpf_error("Failed to lookup cpdom_ctx for %d", i);
 			return -ESRCH;
 		}
-		if (!cpdomc->is_active)
+		if (!cpdomc->is_valid)
 			continue;
 
 		/*
@@ -1936,11 +1486,22 @@ static int init_cpumasks(void)
 	int err = 0;
 
 	bpf_rcu_read_lock();
+	/*
+	 * Allocate active cpumask and initialize it with all online CPUs.
+	 */
 	err = calloc_cpumask(&active_cpumask);
 	active = active_cpumask;
 	if (err || !active)
 		goto out;
 
+	online_cpumask = scx_bpf_get_online_cpumask();
+	nr_cpus_onln = bpf_cpumask_weight(online_cpumask);
+	bpf_cpumask_copy(active, online_cpumask);
+	scx_bpf_put_cpumask(online_cpumask);
+
+	/*
+	 * Allocate the other cpumasks.
+	 */
 	err = calloc_cpumask(&ovrflw_cpumask);
 	if (err)
 		goto out;
@@ -1956,15 +1517,6 @@ static int init_cpumasks(void)
 	err = calloc_cpumask(&little_cpumask);
 	if (err)
 		goto out;
-
-	/*
-	 * Initially activate all online CPUs until we know the system load.
-	 */
-	online_cpumask = scx_bpf_get_online_cpumask();
-	nr_cpus_onln = bpf_cpumask_weight(online_cpumask);
-	bpf_cpumask_copy(active, online_cpumask);
-	scx_bpf_put_cpumask(online_cpumask);
-
 out:
 	bpf_rcu_read_unlock();
 	return err;
@@ -1974,6 +1526,7 @@ static s32 init_per_cpu_ctx(u64 now)
 {
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *turbo, *big, *little, *active, *ovrflw, *cd_cpumask;
+	const struct cpumask *online_cpumask;
 	struct cpdom_ctx *cpdomc;
 	int cpu, i, j, err = 0, nr_cpus_non_zero = 0;
 	u64 cpdom_id;
@@ -1981,6 +1534,7 @@ static s32 init_per_cpu_ctx(u64 now)
 	u16 turbo_cap;
 
 	bpf_rcu_read_lock();
+	online_cpumask = scx_bpf_get_online_cpumask();
 
 	/*
 	 * Prepare cpumasks.
@@ -2015,6 +1569,14 @@ static s32 init_per_cpu_ctx(u64 now)
 		if (err)
 			goto unlock_out;
 
+		err = calloc_cpumask(&cpuc->tmp_l_mask);
+		if (err)
+			goto unlock_out;
+
+		err = calloc_cpumask(&cpuc->tmp_i_mask);
+		if (err)
+			goto unlock_out;
+
 		err = calloc_cpumask(&cpuc->tmp_t_mask);
 		if (err)
 			goto unlock_out;
@@ -2023,13 +1585,18 @@ static s32 init_per_cpu_ctx(u64 now)
 		if (err)
 			goto unlock_out;
 
-		if (bpf_cpumask_test_cpu(cpu, cast_mask(active)))
-			cpu_ctx_init_online(cpuc, cpu, now);
-		else
-			cpu_ctx_init_offline(cpuc, cpu, now);
+		err = calloc_cpumask(&cpuc->tmp_t3_mask);
+		if (err)
+			goto unlock_out;
 
-		cpuc->capacity = get_cpuperf_cap(cpu);
+		cpuc->cpu_id = cpu;
+		cpuc->idle_start_clk = 0;
+		cpuc->lat_cri = 0;
+		cpuc->stopping_tm_est_ns = SCX_SLICE_INF;
+		cpuc->online_clk = now;
 		cpuc->offline_clk = now;
+		cpuc->is_online = bpf_cpumask_test_cpu(cpu, online_cpumask);
+		cpuc->capacity = get_cpuperf_cap(cpu);
 		cpuc->cpdom_poll_pos = cpu % LAVD_CPDOM_MAX_NR;
 		cpuc->min_perf_cri = 1000;
 		cpuc->futex_op = LAVD_FUTEX_OP_INVALID;
@@ -2062,15 +1629,9 @@ static s32 init_per_cpu_ctx(u64 now)
 			nr_cpus_big++;
 			big_capacity += cpuc->capacity;
 			bpf_cpumask_set_cpu(cpu, big);
-			/*
-			 * Initially, all big cores are in the active domain
-			 * and all little cores in the overflow domain.
-			 */
-			bpf_cpumask_set_cpu(cpu, active);
 		}
 		else {
 			bpf_cpumask_set_cpu(cpu, little);
-			bpf_cpumask_set_cpu(cpu, ovrflw);
 			have_little_core = true;
 		}
 
@@ -2097,7 +1658,7 @@ static s32 init_per_cpu_ctx(u64 now)
 			err = -ESRCH;
 			goto unlock_out;
 		}
-		if (!cpdomc->is_active)
+		if (!cpdomc->is_valid)
 			continue;
 
 		bpf_for(i, 0, LAVD_CPU_ID_MAX/64) {
@@ -2105,7 +1666,6 @@ static s32 init_per_cpu_ctx(u64 now)
 			bpf_for(j, 0, 64) {
 				if (cpumask & 0x1LLU << j) {
 			 		cpu = (i * 64) + j;
-					bpf_cpumask_set_cpu(cpu, cd_cpumask);
 					cpuc = get_cpu_ctx_id(cpu);
 					if (!cpuc) {
 						scx_bpf_error("Failed to lookup cpu_ctx: %d", cpu);
@@ -2114,12 +1674,20 @@ static s32 init_per_cpu_ctx(u64 now)
 					}
 					cpuc->cpdom_id = cpdomc->id;
 					cpuc->cpdom_alt_id = cpdomc->alt_id;
+
+					if (bpf_cpumask_test_cpu(cpu, online_cpumask))
+						bpf_cpumask_set_cpu(cpu, cd_cpumask);
+					if (bpf_cpumask_test_cpu(cpu, cast_mask(active)))
+						WRITE_ONCE(cpdomc->is_active, true);
 					cpdomc->nr_cpus++;
 				}
 			}
 		}
 	}
 
+	/*
+	 * Print some useful informatin for debugging.
+	 */
 	bpf_for(cpu, 0, nr_cpu_ids) {
 		cpuc = get_cpu_ctx_id(cpu);
 		if (!cpuc) {
@@ -2132,6 +1700,7 @@ static s32 init_per_cpu_ctx(u64 now)
 	}
 
 unlock_out:
+	scx_bpf_put_cpumask(online_cpumask);
 	bpf_rcu_read_unlock();
 	return err;
 }
@@ -2210,8 +1779,6 @@ SCX_OPS_DEFINE(lavd_ops,
 	       .init_task		= (void *)lavd_init_task,
 	       .init			= (void *)lavd_init,
 	       .exit			= (void *)lavd_exit,
-	       .flags			= SCX_OPS_KEEP_BUILTIN_IDLE |
-					  SCX_OPS_ENQ_EXITING,
 	       .timeout_ms		= 30000U,
 	       .name			= "lavd");
 
