@@ -16,7 +16,6 @@ private(LAVD) struct bpf_cpumask __kptr *big_cpumask; /* CPU mask for big CPUs *
 private(LAVD) struct bpf_cpumask __kptr *little_cpumask; /* CPU mask for little CPUs */
 private(LAVD) struct bpf_cpumask __kptr *active_cpumask; /* CPU mask for active CPUs */
 private(LAVD) struct bpf_cpumask __kptr *ovrflw_cpumask; /* CPU mask for overflow CPUs */
-private(LAVD) struct bpf_cpumask cpdom_cpumask[LAVD_CPDOM_MAX_NR]; /* CPU mask for each compute domain */
 
 const volatile u64	nr_cpu_ids;	/* maximum CPU IDs */
 static volatile u64	nr_cpus_onln;	/* current number of online CPUs */
@@ -34,7 +33,7 @@ volatile bool		no_prefer_turbo_core;
 volatile bool		is_powersave_mode;
 volatile bool		reinit_cpumask_for_performance;
 const volatile bool	is_autopilot_on;
-const volatile u32 	is_smt_active;
+const volatile bool	is_smt_active;
 const volatile u8	verbose;
 
 /*
@@ -140,39 +139,16 @@ static u64 rsigmoid_u64(u64 v, u64 max)
 	return (v >= max) ? 0 : max - v;
 }
 
-static struct task_ctx *try_get_task_ctx(struct task_struct *p)
-{
-	return bpf_task_storage_get(&task_ctx_stor, p, 0, 0);
-}
-
 static struct task_ctx *get_task_ctx(struct task_struct *p)
 {
-	struct task_ctx *taskc;
-
-	taskc = try_get_task_ctx(p);
-	if (!taskc)
-		scx_bpf_error("task_ctx lookup failed for %s[%d]",
-			      p->comm, p->pid);
-	return taskc;
-}
-
-static struct cpu_ctx *try_get_cpu_ctx(void)
-{
-	const u32 idx = 0;
-
-	return bpf_map_lookup_elem(&cpu_ctx_stor, &idx);
+	return bpf_task_storage_get(&task_ctx_stor, p, 0, 0);
 }
 
 static struct cpu_ctx *get_cpu_ctx(void)
 {
 	const u32 idx = 0;
-	struct cpu_ctx *cpuc;
 
-	cpuc = bpf_map_lookup_elem(&cpu_ctx_stor, &idx);
-	if (!cpuc)
-		scx_bpf_error("cpu_ctx lookup failed for current cpu");
-
-	return cpuc;
+	return bpf_map_lookup_elem(&cpu_ctx_stor, &idx);
 }
 
 static struct cpu_ctx *get_cpu_ctx_id(s32 cpu_id)
@@ -231,11 +207,6 @@ static bool is_per_cpu_task(const struct task_struct *p)
 	return false;
 }
 
-static bool have_idle_cpus(const struct cpumask *idle_mask)
-{
-	return !bpf_cpumask_empty(idle_mask);
-}
-
 static bool is_lat_cri(struct task_ctx *taskc, struct sys_stat *stat_cur)
 {
 	return taskc->lat_cri >= stat_cur->avg_lat_cri;
@@ -277,7 +248,7 @@ static bool use_full_cpus(void)
 	return (stat_cur->nr_active + LAVD_CC_NR_OVRFLW) >= nr_cpus_onln;
 }
 
-static u64 pick_any_bit(u64 bitmap, u64 nuance)
+static s64 pick_any_bit(u64 bitmap, u64 nuance)
 {
 	u64 i, pos;
 
