@@ -18,8 +18,6 @@ use std::collections::BTreeMap;
 use std::ffi::c_int;
 use std::ffi::CStr;
 use std::fmt;
-use std::fs::File;
-use std::io::Read;
 use std::mem;
 use std::mem::MaybeUninit;
 use std::str;
@@ -38,22 +36,16 @@ use crossbeam::channel::RecvTimeoutError;
 use crossbeam::channel::Sender;
 use crossbeam::channel::TrySendError;
 use itertools::iproduct;
-use libbpf_rs::skel::OpenSkel;
-use libbpf_rs::skel::Skel;
-use libbpf_rs::skel::SkelBuilder;
 use libbpf_rs::OpenObject;
 use libbpf_rs::ProgramInput;
 use libc::c_char;
 use log::debug;
 use log::info;
-use log::warn;
 use plain::Plain;
 use scx_stats::prelude::*;
 use scx_utils::autopower::{fetch_power_profile, PowerProfile};
 use scx_utils::build_id;
 use scx_utils::compat;
-use scx_utils::import_enums;
-use scx_utils::scx_enums;
 use scx_utils::scx_ops_attach;
 use scx_utils::scx_ops_load;
 use scx_utils::scx_ops_open;
@@ -231,7 +223,6 @@ impl introspec {
 struct CpuFlatId {
     node_id: usize,
     llc_pos: usize,
-    max_freq: usize,
     core_pos: usize,
     cpu_pos: usize,
     cpu_id: usize,
@@ -262,6 +253,7 @@ struct FlatTopology {
     cpu_fids_performance: Vec<CpuFlatId>,
     cpu_fids_powersave: Vec<CpuFlatId>,
     cpdom_map: BTreeMap<ComputeDomainKey, ComputeDomainValue>,
+    smt_enabled: bool,
 }
 
 impl fmt::Display for FlatTopology {
@@ -275,6 +267,7 @@ impl fmt::Display for FlatTopology {
         for (k, v) in self.cpdom_map.iter() {
             write!(f, "\nCPDOM: {:?} {:?}", k, v).ok();
         }
+        write!(f, "SMT: {}", self.smt_enabled).ok();
         Ok(())
     }
 }
@@ -282,43 +275,33 @@ impl fmt::Display for FlatTopology {
 impl FlatTopology {
     /// Build a flat-structured topology
     pub fn new() -> Result<FlatTopology> {
-        let (cpu_fids_performance, avg_freq) = Self::build_cpu_fids(false, false).unwrap();
-        let (cpu_fids_powersave, _) = Self::build_cpu_fids(true, true).unwrap();
+        let topo = Topology::new().expect("Failed to build host topology");
+        let (cpu_fids_performance, avg_cap) = Self::build_cpu_fids(&topo, false, false).unwrap();
+        let (cpu_fids_powersave, _) = Self::build_cpu_fids(&topo, true, true).unwrap();
 
         // Note that building compute domain is not dependent to CPU orer
         // so it is okay to use any cpu_fids_*.
-        let cpdom_map = Self::build_cpdom(&cpu_fids_performance, avg_freq).unwrap();
+        let cpdom_map = Self::build_cpdom(&cpu_fids_performance, avg_cap).unwrap();
 
         Ok(FlatTopology {
             cpu_fids_performance,
             cpu_fids_powersave,
             cpdom_map,
+            smt_enabled: topo.smt_enabled,
         })
     }
 
-    /// Check if SMT is enabled or not
-    pub fn is_smt_active() -> std::io::Result<i32> {
-        let mut file = File::open("/sys/devices/system/cpu/smt/active")?;
-        let mut contents = String::new();
-        file.read_to_string(&mut contents)?;
-
-        let smt_active: i32 = contents.trim().parse().unwrap_or(0);
-
-        Ok(smt_active)
-    }
-
     /// Build a flat-structured list of CPUs in a preference order
     fn build_cpu_fids(
+        topo: &Topology,
         prefer_smt_core: bool,
         prefer_little_core: bool,
     ) -> Option<(Vec<CpuFlatId>, usize)> {
-        let topo = Topology::new().expect("Failed to build host topology");
         let mut cpu_fids = Vec::new();
         debug!("{:#?}", topo);
 
         // Build a vector of cpu flat ids.
-        let mut base_freq = 0;
-        let mut avg_freq = 0;
+        let mut avg_cap = 0;
         for (&node_id, node) in topo.nodes.iter() {
             for (llc_pos, (_llc_id, llc)) in node.llcs.iter().enumerate() {
                 for (core_pos, (core_id, core)) in llc.cores.iter().enumerate() {
@@ -326,7 +309,6 @@ impl FlatTopology {
                         let cpu_fid = CpuFlatId {
                             node_id,
                             llc_pos,
-                            max_freq: cpu.max_freq,
                             core_pos,
                             cpu_pos,
                             cpu_id: *cpu_id,
@@ -334,33 +316,15 @@ impl FlatTopology {
                             l2_id: cpu.l2_id,
                             l3_id: cpu.l3_id,
                             sharing_lvl: 0,
-                            cpu_cap: 0,
+                            cpu_cap: cpu.cpu_capacity,
                         };
                         cpu_fids.push(RefCell::new(cpu_fid));
-                        if base_freq < cpu.max_freq {
-                            base_freq = cpu.max_freq;
-                        }
-                        avg_freq += cpu.max_freq;
+                        avg_cap += cpu.cpu_capacity;
                     }
                 }
             }
         }
-        avg_freq /= cpu_fids.len() as usize;
-
-        // Initialize cpu capacity
-        if base_freq > 0 {
-            for cpu_fid in cpu_fids.iter_mut() {
-                let mut cpu_fid = cpu_fid.borrow_mut();
-                cpu_fid.cpu_cap = ((cpu_fid.max_freq * 1024) / base_freq) as usize;
-            }
-        } else {
-            // Unfortunately, the frequency information in sysfs seems not
-            // always correct in some distributions.
-            for cpu_fid in cpu_fids.iter_mut() {
-                cpu_fid.borrow_mut().cpu_cap = 1024 as usize;
-            }
-            warn!("System does not provide proper CPU frequency information.");
-        }
+        avg_cap /= cpu_fids.len() as usize;
 
         // Initialize cpu's hardware resource sharing level
         for (a_id, cpu_fid_a) in cpu_fids.iter().enumerate() {
@@ -394,71 +358,72 @@ impl FlatTopology {
         // Sort the cpu_fids
         match (prefer_smt_core, prefer_little_core) {
             (true, false) => {
-                // Sort the cpu_fids by node, llc, ^max_freq, ^sharing_lvl, core, and cpu order
+                // Sort the cpu_fids by node, llc, ^cpu_cap, ^sharing_lvl, core, and cpu order
                 cpu_fids.sort_by(|a, b| {
                     a.node_id
                         .cmp(&b.node_id)
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| b.max_freq.cmp(&a.max_freq))
+                        .then_with(|| b.cpu_cap.cmp(&a.cpu_cap))
                         .then_with(|| b.sharing_lvl.cmp(&a.sharing_lvl))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                         .then_with(|| a.cpu_pos.cmp(&b.cpu_pos))
                 });
             }
             (true, true) => {
-                // Sort the cpu_fids by node, llc, max_freq, ^sharing_lvl, core, and cpu order
+                // Sort the cpu_fids by node, llc, cpu_cap, ^sharing_lvl, core, and cpu order
                 cpu_fids.sort_by(|a, b| {
                     a.node_id
                         .cmp(&b.node_id)
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| a.max_freq.cmp(&b.max_freq))
+                        .then_with(|| a.cpu_cap.cmp(&b.cpu_cap))
                         .then_with(|| b.sharing_lvl.cmp(&a.sharing_lvl))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                         .then_with(|| a.cpu_pos.cmp(&b.cpu_pos))
                 });
             }
             (false, false) => {
-                // Sort the cpu_fids by cpu, node, llc, ^max_freq, sharing_lvl, and core order
+                // Sort the cpu_fids by cpu, node, llc, ^cpu_cap, sharing_lvl, and core order
                 cpu_fids.sort_by(|a, b| {
                     a.cpu_pos
                         .cmp(&b.cpu_pos)
                         .then_with(|| a.node_id.cmp(&b.node_id))
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| b.max_freq.cmp(&a.max_freq))
+                        .then_with(|| b.cpu_cap.cmp(&a.cpu_cap))
                         .then_with(|| a.sharing_lvl.cmp(&b.sharing_lvl))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                 });
             }
             (false, true) => {
-                // Sort the cpu_fids by cpu, node, llc, max_freq, sharing_lvl, and core order
+                // Sort the cpu_fids by cpu, node, llc, cpu_cap, sharing_lvl, and core order
                 cpu_fids.sort_by(|a, b| {
                     a.cpu_pos
                         .cmp(&b.cpu_pos)
                         .then_with(|| a.node_id.cmp(&b.node_id))
                         .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| a.max_freq.cmp(&b.max_freq))
+                        .then_with(|| a.cpu_cap.cmp(&b.cpu_cap))
                         .then_with(|| a.sharing_lvl.cmp(&b.sharing_lvl))
                         .then_with(|| a.core_pos.cmp(&b.core_pos))
                 });
             }
         }
 
-        Some((cpu_fids, avg_freq))
+        Some((cpu_fids, avg_cap))
     }
 
     /// Build a list of compute domains
     fn build_cpdom(
         cpu_fids: &Vec<CpuFlatId>,
-        avg_freq: usize,
+        avg_cap: usize,
     ) -> Option<BTreeMap<ComputeDomainKey, ComputeDomainValue>> {
-        // Creat a compute domain map
+        // Creat a compute domain map, where a compute domain is a CPUs that
+        // are under the same node and LLC and have the same core type.
         let mut cpdom_id = 0;
         let mut cpdom_map: BTreeMap<ComputeDomainKey, ComputeDomainValue> = BTreeMap::new();
         for cpu_fid in cpu_fids.iter() {
             let key = ComputeDomainKey {
                 node_id: cpu_fid.node_id,
                 llc_pos: cpu_fid.llc_pos,
-                is_big: cpu_fid.max_freq >= avg_freq,
+                is_big: cpu_fid.cpu_cap >= avg_cap,
             };
             let mut value;
             match cpdom_map.get(&key) {
@@ -479,7 +444,9 @@ impl FlatTopology {
             cpdom_map.insert(key, value);
         }
 
-        // Fill up cpdom_alt_id for each compute domain
+        // Fill up cpdom_alt_id for each compute domain, where the alternative
+        // compute domain is a compute domain that are under the same node
+        // and LLC but has a different core type.
         for (k, v) in cpdom_map.iter() {
             let mut key = k.clone();
             key.is_big = !k.is_big;
@@ -489,7 +456,8 @@ impl FlatTopology {
             }
         }
 
-        // Build a neighbor map for each compute domain
+        // Build a neighbor map for each compute domain, where neighbors are
+        // ordered by core type, node, and LLC.
         for ((from_k, from_v), (to_k, to_v)) in iproduct!(cpdom_map.iter(), cpdom_map.iter()) {
             if from_k == to_k {
                 continue;
@@ -587,7 +555,7 @@ impl<'a> Scheduler<'a> {
         Self::init_cpus(&mut skel, &topo);
 
         // Initialize skel according to @opts.
-        Self::init_globals(&mut skel, &opts);
+        Self::init_globals(&mut skel, &opts, &topo);
 
         // Attach.
         let mut skel = scx_ops_load!(skel, lavd_ops, uei)?;
@@ -635,7 +603,7 @@ impl<'a> Scheduler<'a> {
             skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].alt_id = v.cpdom_alt_id.get() as u64;
             skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].node_id = k.node_id as u8;
             skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].is_big = k.is_big as u8;
-            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].is_active = 1;
+            skel.maps.bss_data.cpdom_ctxs[v.cpdom_id].is_valid = 1;
             for cpu_id in v.cpu_ids.iter() {
                 let i = cpu_id / 64;
                 let j = cpu_id % 64;
@@ -663,20 +631,23 @@ impl<'a> Scheduler<'a> {
         opts.prefer_smt_core && opts.prefer_little_core
     }
 
-    fn init_globals(skel: &mut OpenBpfSkel, opts: &Opts) {
+    fn init_globals(skel: &mut OpenBpfSkel, opts: &Opts, topo: &FlatTopology) {
         skel.maps.bss_data.no_core_compaction = opts.no_core_compaction;
         skel.maps.bss_data.no_freq_scaling = opts.no_freq_scaling;
         skel.maps.bss_data.no_prefer_turbo_core = opts.no_prefer_turbo_core;
         skel.maps.bss_data.is_powersave_mode = Self::is_powersave_mode(&opts);
         skel.maps.rodata_data.nr_cpu_ids = *NR_CPU_IDS as u64;
-        skel.maps.rodata_data.is_smt_active = match FlatTopology::is_smt_active() {
-            Ok(ret) => (ret == 1) as u32,
-            Err(_) => 0,
-        };
+        skel.maps.rodata_data.is_smt_active = topo.smt_enabled;
         skel.maps.rodata_data.is_autopilot_on = opts.autopilot;
         skel.maps.rodata_data.verbose = opts.verbose;
         skel.maps.rodata_data.slice_max_ns = opts.slice_max_us * 1000;
         skel.maps.rodata_data.slice_min_ns = opts.slice_min_us * 1000;
+
+        skel.struct_ops.lavd_ops_mut().flags = *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP
+            | *compat::SCX_OPS_ENQ_EXITING
+            | *compat::SCX_OPS_ENQ_LAST
+            | *compat::SCX_OPS_ENQ_MIGRATION_DISABLED
+            | *compat::SCX_OPS_KEEP_BUILTIN_IDLE;
     }
 
     fn get_msg_seq_id() -> u64 {
