@@ -111,11 +111,11 @@ enum layer_stat_id {
 	LSTAT_ENQ_WAKEUP,
 	LSTAT_ENQ_EXPIRE,
 	LSTAT_ENQ_REENQ,
+	LSTAT_KEEP,
 	LSTAT_MIN_EXEC,
 	LSTAT_MIN_EXEC_NS,
 	LSTAT_OPEN_IDLE,
 	LSTAT_AFFN_VIOL,
-	LSTAT_KEEP,
 	LSTAT_KEEP_FAIL_MAX_EXEC,
 	LSTAT_KEEP_FAIL_BUSY,
 	LSTAT_PREEMPT,
@@ -169,7 +169,8 @@ struct cpu_ctx {
 	bool			running_owned;
 	bool			running_open;
 	bool			running_fallback;
-	u64			running_at;
+	u64			used_at;
+	bool			is_protected;
 
 	u64			layer_usages[MAX_LAYERS][NR_LAYER_USAGES];
 	u64			gstats[NR_GSTATS];
@@ -267,6 +268,7 @@ struct layer_match {
 	bool		is_kthread;
 	bool		used_gpu_tid;
 	bool		used_gpu_pid;
+	bool		exclude;
 };
 
 struct layer_match_ands {
@@ -283,7 +285,11 @@ enum layer_growth_algo {
 	GROWTH_ALGO_ROUND_ROBIN,
 	GROWTH_ALGO_BIG_LITTLE,
 	GROWTH_ALGO_LITTLE_BIG,
+	GROWTH_ALGO_NODE_SPREAD,
+	GROWTH_ALGO_NODE_SPREAD_REVERSE,
+	GROWTH_ALGO_NODE_SPREAD_RANDOM,
 	GROWTH_ALGO_RANDOM_TOPO,
+	GROWTH_ALGO_STICKY_DYNAMIC,
 };
 
 struct layer {
@@ -305,6 +311,7 @@ struct layer {
 	bool			preempt_first;
 	bool			exclusive;
 	bool			allow_node_aligned;
+	bool			prev_over_idle_core;
 	int			growth_algo;
 
 	u64			nr_tasks;
@@ -324,6 +331,7 @@ struct layer {
 	u32			llc_drain_cnt;
 
 	char			name[MAX_LAYER_NAME];
+	bool			is_protected;
 };
 
 struct scx_cmd {
