@@ -62,7 +62,9 @@ impl LayerSpec {
 pub enum LayerMatch {
     CgroupPrefix(String),
     CommPrefix(String),
+    CommPrefixExclude(String),
     PcommPrefix(String),
+    PcommPrefixExclude(String),
     NiceAbove(i32),
     NiceBelow(i32),
     NiceEquals(i32),
@@ -99,6 +101,8 @@ pub struct LayerCommon {
     #[serde(default)]
     pub allow_node_aligned: bool,
     #[serde(default)]
+    pub prev_over_idle_core: bool,
+    #[serde(default)]
     pub weight: u32,
     #[serde(default)]
     pub disallow_open_after_us: Option<u64>,
@@ -113,6 +117,8 @@ pub struct LayerCommon {
     #[serde(default)]
     pub perf: u64,
     #[serde(default)]
+    pub idle_resume_us: Option<u32>,
+    #[serde(default)]
     pub nodes: Vec<usize>,
     #[serde(default)]
     pub llcs: Vec<usize>,
@@ -125,6 +131,12 @@ pub enum LayerKind {
         #[serde(default)]
         cpus_range: Option<(usize, usize)>,
 
+        #[serde(default)]
+        cpus_range_frac: Option<(f64, f64)>,
+
+        #[serde(default)]
+        protected: bool,
+
         #[serde(flatten)]
         common: LayerCommon,
     },
@@ -133,6 +145,12 @@ pub enum LayerKind {
         #[serde(default)]
         cpus_range: Option<(usize, usize)>,
 
+        #[serde(default)]
+        cpus_range_frac: Option<(f64, f64)>,
+
+        #[serde(default)]
+        protected: bool,
+
         #[serde(flatten)]
         common: LayerCommon,
     },
