@@ -25,9 +25,6 @@ use bitvec::prelude::*;
 use cgroupfs::CgroupReader;
 use clap::Parser;
 use itertools::Itertools;
-use libbpf_rs::skel::OpenSkel;
-use libbpf_rs::skel::Skel;
-use libbpf_rs::skel::SkelBuilder;
 use libbpf_rs::MapCore as _;
 use libbpf_rs::OpenObject;
 use log::debug;
@@ -36,7 +33,6 @@ use log::trace;
 use maplit::btreemap;
 use maplit::hashmap;
 use scx_utils::compat;
-use scx_utils::import_enums;
 use scx_utils::init_libbpf_logging;
 use scx_utils::ravg::ravg_read;
 use scx_utils::scx_enums;
