@@ -2,8 +2,8 @@
 
 // This software may be used and distributed according to the terms of the
 // GNU General Public License version 2.
-#ifndef __INTF_H
-#define __INTF_H
+#ifndef __P2DQ_INTF_H
+#define __P2DQ_INTF_H
 
 #include <stdbool.h>
 #ifndef __kptr
@@ -51,7 +51,11 @@ enum stat_idx {
 	P2DQ_STAT_LLC_MIGRATION,
 	P2DQ_STAT_NODE_MIGRATION,
 	P2DQ_STAT_KEEP,
-	P2DQ_STAT_PICK2,
+	P2DQ_STAT_SELECT_PICK2,
+	P2DQ_STAT_DISPATCH_PICK2,
+	P2DQ_STAT_WAKE_PREV,
+	P2DQ_STAT_WAKE_LLC,
+	P2DQ_STAT_WAKE_MIG,
 	P2DQ_NR_STATS,
 };
 
@@ -77,13 +81,15 @@ struct task_ctx {
 	u64			last_dsq_id;
 	int			last_dsq_index;
 	u64 			last_run_at;
+	u64			llc_runs; /* how many runs on the current LLC */
 
 	/* The task is a workqueue worker thread */
 	bool			is_kworker;
 
-	/* Allowed on all CPUs and eligible for DIRECT_GREEDY optimization */
+	/* Allowed to run on all CPUs */
 	bool			all_cpus;
 
+
 	struct bpf_cpumask __kptr *mask;
 };
 
@@ -111,6 +117,7 @@ struct llc_ctx {
 	u64				dsq_load[MAX_DSQS_PER_LLC];
 	struct bpf_cpumask __kptr	*cpumask;
 	struct bpf_cpumask __kptr	*big_cpumask;
+	struct bpf_cpumask __kptr	*tmp_cpumask;
 };
 
 struct node_ctx {
@@ -120,4 +127,31 @@ struct node_ctx {
 	struct bpf_cpumask __kptr	*big_cpumask;
 };
 
-#endif /* __INTF_H */
+enum enqueue_promise_kind {
+	P2DQ_ENQUEUE_PROMISE_COMPLETE,
+	P2DQ_ENQUEUE_PROMISE_VTIME,
+	P2DQ_ENQUEUE_PROMISE_FIFO,
+};
+
+struct enqueue_promise_vtime {
+	u64	dsq_id;
+	u64	enq_flags;
+	u64	slice_ns;
+	u64	vtime;
+};
+
+struct enqueue_promise_fifo {
+	u64	dsq_id;
+	u64	enq_flags;
+	u64	slice_ns;
+};
+
+struct enqueue_promise {
+	enum enqueue_promise_kind	kind;
+	union {
+		struct enqueue_promise_vtime	vtime;
+		struct enqueue_promise_fifo	fifo;
+	};
+};
+
+#endif /* __P2DQ_INTF_H */
