@@ -14,6 +14,10 @@
 #include <scx/common.bpf.h>
 #endif
 
+#ifndef P2DQ_CREATE_STRUCT_OPS
+#define P2DQ_CREATE_STRUCT_OPS 1
+#endif
+
 #include "intf.h"
 
 #include <errno.h>
@@ -40,16 +44,22 @@ const volatile u32 nr_dsqs_per_llc = 3;
 const volatile u64 dsq_shift = 2;
 const volatile int init_dsq_index = 0;
 const volatile u64 min_slice_us = 100;
+const volatile u64 min_llc_runs_pick2 = 5;
 const volatile u32 interactive_ratio = 10;
+const volatile u32 min_nr_queued_pick2 = 10;
 
 const volatile bool autoslice = true;
+const volatile bool dispatch_pick2_disable = false;
+const volatile bool eager_load_balance = true;
+const volatile bool greedy_idle = false;
 const volatile bool interactive_sticky = false;
 const volatile bool keep_running_enabled = true;
-const volatile bool eager_load_balance = true;
+const volatile bool kthreads_local = true;
+const volatile bool max_dsq_pick2 = false;
+const volatile bool select_idle_in_enqueue = true;
+
 const volatile bool smt_enabled = true;
-const volatile bool greedy_idle = false;
 const volatile bool has_little_cores = false;
-const volatile bool kthreads_local;
 const volatile u32 debug = 2;
 
 const u32 zero_u32 = 0;
@@ -123,7 +133,10 @@ static struct cpu_ctx *lookup_cpu_ctx(int cpu)
 }
 
 static __always_inline u64 cpu_dsq_id(int dsq_index, struct cpu_ctx *cpuc) {
-	if (dsq_index < 0 || dsq_index > nr_dsqs_per_llc || dsq_index >= MAX_DSQS_PER_LLC) {
+	if (!cpuc ||
+	    dsq_index < 0 ||
+	    dsq_index > nr_dsqs_per_llc ||
+	    dsq_index >= MAX_DSQS_PER_LLC) {
 		scx_bpf_error("cpuc invalid dsq index: %d", dsq_index);
 		return 0;
 	}
@@ -227,6 +240,20 @@ static bool is_interactive(struct task_ctx *taskc)
 	return taskc->dsq_index == 0;
 }
 
+/*
+ * Returns if the task is able to load balance using pick2.
+ */
+static bool can_pick2(struct task_ctx *taskc)
+{
+	if (is_interactive(taskc) ||
+	    !taskc->all_cpus ||
+	    taskc->is_kworker ||
+	    (max_dsq_pick2 > 0 && taskc->llc_runs < min_llc_runs_pick2))
+		return false;
+
+	return true;
+}
+
 /*
  * Returns a random llc_ctx
  */
@@ -266,19 +293,23 @@ static bool keep_running(struct cpu_ctx *cpuc, struct task_struct *p)
 	return true;
 }
 
-static struct llc_ctx *pick_two_llc_ctx(struct llc_ctx *left, struct llc_ctx *right, bool most_loaded)
+static struct llc_ctx *pick_two_llc_ctx(struct llc_ctx *cur_llcx, struct llc_ctx *left,
+					struct llc_ctx *right)
 {
-	s32 left_queued = 0, right_queued = 0;
-	u64 left_load = 0, right_load = 0;
+	s32 cur_queued = 0, left_queued = 0, right_queued = 0;
+	u64 cur_load = 0, left_load = 0, right_load = 0;
 	int i;
 
 	if (!left || !right)
-		return left;
+		return NULL;
 
 	bpf_for(i, 0, nr_llcs) {
 		if (i >= nr_dsqs_per_llc || i < 0)
 			continue;
 
+		u64 cur_dsq_id = *MEMBER_VPTR(cur_llcx->dsqs, [i]);
+		cur_queued += scx_bpf_dsq_nr_queued(cur_dsq_id);
+		cur_load += *MEMBER_VPTR(cur_llcx->dsq_load, [i]);
 		u64 left_dsq_id = *MEMBER_VPTR(left->dsqs, [i]);
 		left_queued += scx_bpf_dsq_nr_queued(left_dsq_id);
 		left_load += *MEMBER_VPTR(left->dsq_load, [i]);
@@ -287,21 +318,42 @@ static struct llc_ctx *pick_two_llc_ctx(struct llc_ctx *left, struct llc_ctx *ri
 		right_load += *MEMBER_VPTR(right->dsq_load, [i]);
 	}
 
-	if (((left_queued > 0 || right_queued > 0) && left_queued < right_queued) ||
-	    left_load > right_load)
-		return most_loaded ? right : left;
+	if (min_nr_queued_pick2 > 0 &&
+	    (left_queued < min_nr_queued_pick2 &&
+	     right_queued < min_nr_queued_pick2))
+		return NULL;
 
-	return most_loaded ? left : right;
+	// If the current LLCs has more load don't try to pick2.
+	if ((nr_llcs > 2 && cur_queued >= left_queued && cur_queued >= right_queued) ||
+	    (nr_llcs > 2 && cur_load > left_load && cur_load > right_load))
+	    return NULL;
+
+        if (left_queued < right_queued || left_load < right_load)
+		return right;
+	return left;
 }
 
-static s32 pick_two_cpu(struct task_ctx *taskc, bool *is_idle)
+static s32 pick_two_cpu(struct llc_ctx *cur_llcx, struct task_ctx *taskc,
+			bool *is_idle)
 {
+	if ((min_llc_runs_pick2 > 0 &&
+	     taskc->llc_runs < min_llc_runs_pick2) ||
+	    !can_pick2(taskc))
+		return -EINVAL;
+
 	struct llc_ctx *chosen;
-	struct llc_ctx *left = rand_llc_ctx();
-	struct llc_ctx *right = rand_llc_ctx();
+	struct llc_ctx *left, *right;
 	int dsq_index = taskc->dsq_index;
 	s32 cpu;
 
+	if (nr_llcs == 2) {
+		left = lookup_llc_ctx(0);
+		right = lookup_llc_ctx(1);
+	} else {
+		left = rand_llc_ctx();
+		right = rand_llc_ctx();
+	}
+
 	if (!left || !right || dsq_index > nr_dsqs_per_llc) {
 		return -EINVAL;
 	}
@@ -313,6 +365,9 @@ static s32 pick_two_cpu(struct task_ctx *taskc, bool *is_idle)
 			return -EINVAL;
 	}
 
+	u64 cur_dsq_id = *MEMBER_VPTR(cur_llcx->dsqs, [dsq_index]);
+	s32 cur_queued = scx_bpf_dsq_nr_queued(cur_dsq_id);
+	u64 cur_load = *MEMBER_VPTR(cur_llcx->dsq_load, [dsq_index]);
 	u64 left_dsq_id = *MEMBER_VPTR(left->dsqs, [dsq_index]);
 	s32 left_queued = scx_bpf_dsq_nr_queued(left_dsq_id);
 	u64 left_load = *MEMBER_VPTR(left->dsq_load, [dsq_index]);
@@ -320,10 +375,19 @@ static s32 pick_two_cpu(struct task_ctx *taskc, bool *is_idle)
 	s32 right_queued = scx_bpf_dsq_nr_queued(right_dsq_id);
 	u64 right_load = *MEMBER_VPTR(right->dsq_load, [dsq_index]);
 
-	stat_inc(P2DQ_STAT_PICK2);
+	// If the other LLCs have more load than the current don't bother.
+	if ((nr_llcs > 2 && left_queued >= cur_queued && right_queued >= cur_queued) ||
+	    (nr_llcs > 2 && left_load > cur_load && right_load > cur_load))
+		return -EINVAL;
 
-	if (((left_queued > 0 || right_queued > 0) && left_queued < right_queued) ||
-	    left_load > right_load) {
+	if (min_nr_queued_pick2 > 0 &&
+	    (left_queued < min_nr_queued_pick2 &&
+	     right_queued < min_nr_queued_pick2))
+		return -EINVAL;
+
+	if (((left_queued > 0 || right_queued > 0) ||
+	    left_queued < right_queued) ||
+	    left_load < right_load) {
 		chosen = left;
 		goto pick_llc;
 	} else {
@@ -332,24 +396,28 @@ static s32 pick_two_cpu(struct task_ctx *taskc, bool *is_idle)
 	}
 
 pick_llc:
-	if (!chosen)
+	if (!chosen || !chosen->cpumask)
 		return -EINVAL;
 
-	if (!chosen->cpumask)
-		return -EINVAL;
-
-	cpu = scx_bpf_pick_idle_cpu(cast_mask(chosen->cpumask), 0);
-	if (cpu < nr_cpus && cpu >= 0) {
+	// First try to find an idle core
+	cpu = scx_bpf_pick_idle_cpu(cast_mask(chosen->cpumask), SCX_PICK_IDLE_CORE);
+	if (cpu >= 0) {
 		*is_idle = true;
 		return cpu;
 	}
 
-	// couldn't find idle, but still return a CPU.
-	if (!chosen->cpumask)
-		return -EINVAL;
-	cpu = bpf_cpumask_any_distribute(cast_mask(chosen->cpumask));
-	if (cpu < nr_cpus)
+	// No idle cores, any CPU will do
+	if (chosen->cpumask &&
+	    (cpu = scx_bpf_pick_idle_cpu(cast_mask(chosen->cpumask), 0)) >= 0) {
+		*is_idle = true;
 		return cpu;
+	}
+
+	// Couldn't find idle, but still return a CPU to load balance
+	if (chosen->cpumask &&
+	    (cpu = bpf_cpumask_any_distribute(cast_mask(chosen->cpumask))) < nr_cpus) {
+		return cpu;
+	}
 
 	return -EINVAL;
 }
@@ -361,25 +429,76 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 	struct cpu_ctx *prev_cpuc;
 	struct node_ctx *nodec;
 	struct llc_ctx *llcx;
-	bool is_prev_llc_affine = false;
 	bool interactive = is_interactive(taskc);
+	int llc_idle = 0;
 	s32 cpu = prev_cpu;
 
 	idle_cpumask = scx_bpf_get_idle_cpumask();
 	idle_smtmask = scx_bpf_get_idle_smtmask();
 
-	if (!idle_cpumask || !idle_smtmask || bpf_cpumask_empty(idle_cpumask))
-		goto out_put_cpumask;
+	if (!idle_cpumask || !idle_smtmask)
+		goto found_cpu;
 
 	if (!(prev_cpuc = lookup_cpu_ctx(prev_cpu)) ||
 	    !(llcx = lookup_llc_ctx(prev_cpuc->llc_id)) ||
 	    !(nodec = lookup_node_ctx(prev_cpuc->node_id)) ||
-	    !llcx->cpumask)
-		goto out_put_cpumask;
+	    !llcx->cpumask || !llcx->tmp_cpumask)
+		goto found_cpu;
+
+	bpf_cpumask_and(llcx->tmp_cpumask, idle_cpumask, cast_mask(llcx->cpumask));
+	int nr_idle = bpf_cpumask_weight(idle_cpumask);
+	int nr_queued = scx_bpf_dsq_nr_queued(taskc->dsq_id);
+	if (llcx->tmp_cpumask)
+		llc_idle = bpf_cpumask_weight(cast_mask(llcx->tmp_cpumask));
+
+
+	// Special handling of tasks with custom affinities
+	if (!taskc->all_cpus) {
+		// First try last CPU
+		if (bpf_cpumask_test_cpu(prev_cpu, p->cpus_ptr)) {
+			cpu = prev_cpu;
+			goto found_cpu;
+		}
+
+		if (taskc->mask && llcx->cpumask)
+			bpf_cpumask_and(taskc->mask, cast_mask(llcx->cpumask),
+					p->cpus_ptr);
+
+		// First try to find an idle SMT in the LLC
+		if (smt_enabled && taskc->mask) {
+			cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask),
+						    SCX_PICK_IDLE_CORE);
+			if (cpu >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+
+		// Next try to find an idle CPU in the LLC
+		if (taskc->mask) {
+			cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask), 0);
+			if (cpu >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+
+		// Next try to find an idle CPU in the node
+		if (nodec->cpumask && taskc->mask) {
+			bpf_cpumask_and(taskc->mask, cast_mask(nodec->cpumask),
+					p->cpus_ptr);
+			if (taskc->mask &&
+			    (cpu = scx_bpf_pick_idle_cpu(cast_mask(taskc->mask), 0)) >= 0) {
+				*is_idle = true;
+				goto found_cpu;
+			}
+		}
+
+		// Fallback to anywhere the task can run
+		cpu = bpf_cpumask_any_distribute(p->cpus_ptr);
+		goto found_cpu;
+	}
 
-	is_prev_llc_affine = bpf_cpumask_test_cpu(prev_cpu, cast_mask(llcx->cpumask));
-	if (!llcx->cpumask || !idle_cpumask)
-		goto out_put_cpumask;
 
 	/*
 	 * If the current task is waking up another task and releasing the CPU
@@ -388,136 +507,114 @@ static s32 pick_idle_cpu(struct task_struct *p, struct task_ctx *taskc,
 	 */
 	if (wake_flags & SCX_WAKE_SYNC) {
 		struct task_struct *current = (void *)bpf_get_current_task_btf();
-		if (!current)
-			goto out_put_cpumask;
-
 		struct task_ctx *cur_taskc = lookup_task_ctx_may_fail(current);
-		if (!cur_taskc)
-			goto out_put_cpumask;
+		if (cur_taskc) {
+			if (cur_taskc->llc_id == llcx->id) {
+				// First check if the waking task is in the same LLC
+				// and the prev cpu is idle
+				if (scx_bpf_test_and_clear_cpu_idle(cpu)) {
+					stat_inc(P2DQ_STAT_WAKE_PREV);
+					*is_idle = true;
+					goto found_cpu;
+				}
+				// Try an idle core in the LLC.
+				if (llcx->cpumask &&
+				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
+								 SCX_PICK_IDLE_CORE)) >= 0) {
+					stat_inc(P2DQ_STAT_WAKE_LLC);
+					*is_idle = true;
+					goto found_cpu;
+				}
+			} else {
+				// If not in the local LLC we need to migrate to the waker llc.
+				struct llc_ctx *cur_llcx = lookup_llc_ctx(cur_taskc->llc_id);
+				if (!cur_llcx)
+					goto found_cpu;
 
-		// If the waking task is in the same LLC then use the current
-		// CPU if available.
-		s32 cur_cpu = bpf_get_smp_processor_id();
-		if (llcx->id == cur_taskc->llc_id &&
-		    scx_bpf_test_and_clear_cpu_idle(cur_cpu)) {
-			*is_idle = true;
-			goto out_put_cpumask;
+				if (cur_llcx->cpumask &&
+				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
+								 SCX_PICK_IDLE_CORE)) >= 0) {
+					stat_inc(P2DQ_STAT_WAKE_MIG);
+					*is_idle = true;
+					goto found_cpu;
+				}
+
+				// Couldn't find an idle core so just migrate to the CPU
+				if (cur_llcx->cpumask &&
+				    (cpu = scx_bpf_pick_idle_cpu(cast_mask(cur_llcx->cpumask),
+								 0)) >= 0) {
+					stat_inc(P2DQ_STAT_WAKE_MIG);
+					*is_idle = true;
+					goto found_cpu;
+				}
+			}
 		}
 	}
 
-	// If last CPU is idle then run again
-	if (is_prev_llc_affine &&
-	    bpf_cpumask_test_cpu(prev_cpu, smt_enabled ? idle_smtmask : idle_cpumask) &&
-	    scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
-		cpu = prev_cpu;
-		*is_idle = true;
-		goto out_put_cpumask;
-	}
-
-	if (eager_load_balance && nr_llcs > 1) {
-		llcx = pick_two_llc_ctx(rand_llc_ctx(), rand_llc_ctx(), false);
-		if (!llcx) {
-			cpu = prev_cpu;
-			goto out_put_cpumask;
+	if (nr_llcs > 1 &&
+	    (eager_load_balance ||
+	     (llc_idle == 0 && nr_queued > llcx->nr_cpus && nr_idle > 0))) {
+		cpu = pick_two_cpu(llcx, taskc, is_idle);
+		if (cpu >= 0) {
+			stat_inc(P2DQ_STAT_SELECT_PICK2);
+			goto found_cpu;
 		}
 	}
 
 	if (has_little_cores) {
-		if (!llcx->big_cpumask)
-			goto out_put_cpumask;
-
-		cpu = bpf_cpumask_any_and_distribute(smt_enabled ? idle_smtmask : idle_cpumask,
-						     cast_mask(llcx->big_cpumask));
-		if (cpu < nr_cpus) {
+		if (llcx->big_cpumask &&
+		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->big_cpumask),
+						 SCX_PICK_IDLE_CORE)) >= 0) {
 			*is_idle = true;
-			goto out_put_cpumask;
+			goto found_cpu;
 		}
 
-		if (!nodec->big_cpumask)
-			goto out_put_cpumask;
-
-		/*
-		 * Next try a big core in the local node.
-		 */
-		if (!interactive) {
-			cpu = bpf_cpumask_any_and_distribute(smt_enabled ? idle_smtmask : idle_cpumask,
-							     cast_mask(nodec->big_cpumask));
-			if (cpu < nr_cpus) {
-				*is_idle = true;
-				goto out_put_cpumask;
-			}
-		}
-	}
-
-	if (smt_enabled) {
-		if (is_prev_llc_affine &&
-		    bpf_cpumask_test_cpu(prev_cpu, idle_smtmask) &&
-		    scx_bpf_test_and_clear_cpu_idle(prev_cpu)) {
-			cpu = prev_cpu;
+		if (nodec->big_cpumask &&
+		    (cpu = scx_bpf_pick_idle_cpu(cast_mask(nodec->big_cpumask),
+						 SCX_PICK_IDLE_CORE)) >= 0) {
 			*is_idle = true;
-			goto out_put_cpumask;
+			goto found_cpu;
 		}
 	}
 
-	if (!llcx->cpumask)
-		goto out_put_cpumask;
-
-	// Next try in the local LLC.
-	cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask), SCX_PICK_IDLE_CORE);
-	if (cpu >= 0) {
-		*is_idle = true;
-		goto out_put_cpumask;
-	}
-
-	if (interactive_sticky && interactive) {
+	// First check if last CPU is idle
+	if (llcx->cpumask &&
+	    bpf_cpumask_test_cpu(prev_cpu, cast_mask(llcx->cpumask)) &&
+	    bpf_cpumask_test_cpu(prev_cpu, smt_enabled ? idle_smtmask : idle_cpumask)) {
 		cpu = prev_cpu;
-		goto out_put_cpumask;
-	}
-
-	// Greedy idle tries to grab any idle CPU in the LLC
-	if (greedy_idle && llcx->cpumask) {
-		cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask), 0);
-		if (cpu >= 0) {
-			*is_idle = true;
-			stat_inc(P2DQ_STAT_GREEDY_IDLE);
-			goto out_put_cpumask;
-		}
-	}
-
-	if (!nodec->cpumask)
-		goto out_put_cpumask;
-
-	cpu = scx_bpf_pick_idle_cpu(cast_mask(nodec->cpumask), SCX_PICK_IDLE_CORE);
-	if (cpu >= 0) {
 		*is_idle = true;
-		goto out_put_cpumask;
+		goto found_cpu;
 	}
 
-	if (nr_llcs > 1) {
-		// Try to keep interactive tasks local to the LLC and let non
-		// interactive tasks load balance via pick2.
-		if (interactive && nodec->cpumask) {
-			cpu = scx_bpf_pick_idle_cpu(cast_mask(nodec->cpumask), 0);
-			if (cpu >= 0) {
-				*is_idle = true;
-				goto out_put_cpumask;
-			}
-		} else {
-			cpu = pick_two_cpu(taskc, is_idle);
-			if (cpu >= 0) {
-				goto out_put_cpumask;
-			}
-		}
-	}
-
-	cpu = scx_bpf_pick_idle_cpu(p->cpus_ptr, 0);
-	if (cpu >= 0) {
+	// Next try in the local LLC
+	if (llcx->cpumask &&
+	    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask),
+					 SCX_PICK_IDLE_CORE)) >= 0) {
 		*is_idle = true;
-		goto out_put_cpumask;
+		goto found_cpu;
 	}
-	cpu = prev_cpu;
 
-out_put_cpumask:
+	// Try a idle CPU in the llc
+	if (llcx->cpumask &&
+	    (cpu = scx_bpf_pick_idle_cpu(cast_mask(llcx->cpumask), 0)) >= 0) {
+		*is_idle = true;
+		goto found_cpu;
+	}
+
+	// Non-interactive tasks load balance
+	if (nr_llcs > 1 &&
+	    !interactive &&
+	    nr_idle > 1 &&
+	    (cpu = pick_two_cpu(llcx, taskc, is_idle)) >= 0) {
+		stat_inc(P2DQ_STAT_SELECT_PICK2);
+		goto found_cpu;
+	}
+
+	// Couldn't find anything idle just return something in the local LLC
+	if (llcx->cpumask)
+		cpu = bpf_cpumask_any_distribute(cast_mask(llcx->cpumask));
+
+found_cpu:
 	scx_bpf_put_cpumask(idle_cpumask);
 	scx_bpf_put_cpumask(idle_smtmask);
 	if (cpu >= nr_cpus || cpu < 0)
@@ -527,7 +624,7 @@ out_put_cpumask:
 }
 
 
-s32 BPF_STRUCT_OPS(p2dq_select_cpu, struct task_struct *p, s32 prev_cpu, u64 wake_flags)
+static __always_inline s32 p2dq_select_cpu_impl(struct task_struct *p, s32 prev_cpu, u64 wake_flags)
 {
 	struct task_ctx *taskc;
 	bool is_idle = false;
@@ -547,9 +644,20 @@ s32 BPF_STRUCT_OPS(p2dq_select_cpu, struct task_struct *p, s32 prev_cpu, u64 wak
 }
 
 
-void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_flags)
+/*
+ * Perform the enqueue logic for `p` but don't enqueue it where possible.  This
+ * is primarily used so that scx_chaos can decide to enqueue a task either
+ * immediately in `enqueue` or later in `dispatch`. This returns a tagged union
+ * with three states:
+ * - P2DQ_ENQUEUE_PROMISE_COMPLETE: Either the task has been enqueued, or there
+ *     is nothing to do (enqueue failed).
+ * - P2DQ_ENQUEUE_PROMISE_FIFO: The completer should enqueue this task on a fifo dsq.
+ * - P2DQ_ENQUEUE_PROMISE_VTIME: The completer should enqueue this task on a vtime dsq.
+ */
+static __always_inline void async_p2dq_enqueue(struct enqueue_promise *ret,
+					       struct task_struct *p,
+					       u64 enq_flags)
 {
-	const struct cpumask *idle_cpumask, *llc_mask;
 	struct llc_ctx *llcx, *prev_llcx;
 	struct cpu_ctx *cpuc, *task_cpuc;
 	struct task_ctx *taskc;
@@ -560,20 +668,25 @@ void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_f
 	if (!(cpuc = lookup_cpu_ctx(-1)) ||
 	    !(task_cpuc = lookup_cpu_ctx(task_cpu)) ||
 	    !(taskc = lookup_task_ctx(p)) ||
-	    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
+	    !(llcx = lookup_llc_ctx(cpuc->llc_id))) {
+		ret->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 		return;
+	}
 
 	u64 vtime_now = llcx->vtime;
 	u64 slice_ns = dsq_time_slice(taskc->dsq_index);
 
 	// If the task in in another LLC need to update vtime.
 	if (taskc->llc_id != cpuc->llc_id) {
-		if (!(prev_llcx = lookup_llc_ctx(task_cpuc->llc_id)))
+		if (!(prev_llcx = lookup_llc_ctx(task_cpuc->llc_id))) {
+			ret->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 			return;
+		}
 
 		u64 vtime_delta = p->scx.dsq_vtime - prev_llcx->vtime;
 		p->scx.dsq_vtime = vtime_now + vtime_delta;
-		trace("vtime change %llu, new vtime %llu", vtime_delta, p->scx.dsq_vtime);
+		trace("vtime change %llu, new vtime %llu",
+		      vtime_delta, p->scx.dsq_vtime);
 	}
 
 	u64 vtime = p->scx.dsq_vtime;
@@ -598,50 +711,72 @@ void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_f
 		stat_inc(P2DQ_STAT_DIRECT);
 		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_ns,
 				   enq_flags | SCX_ENQ_PREEMPT);
+
+		ret->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
 		return;
 	}
 
-	/* 
+	/*
 	 * Affinitized tasks just get dispatched directly, need to handle this
 	 * better.
 	 */
-	if ((!taskc->all_cpus)) {
+	if (!taskc->all_cpus && bpf_cpumask_test_cpu(cpuc->id, p->cpus_ptr)) {
 		stat_inc(P2DQ_STAT_DIRECT);
-		scx_bpf_dsq_insert(p, SCX_DSQ_LOCAL, slice_ns, enq_flags);
+
+		ret->kind = P2DQ_ENQUEUE_PROMISE_FIFO;
+		ret->fifo.dsq_id = SCX_DSQ_LOCAL;
+		ret->fifo.enq_flags = enq_flags;
+		ret->fifo.slice_ns = slice_ns;
 		return;
 	}
 
+	// If an idle CPU hasn't been found in select_cpu find one now
+	if (select_idle_in_enqueue && !__COMPAT_is_enq_cpu_selected(enq_flags)) {
+		bool is_idle = false;
+		cpu = pick_idle_cpu(p, taskc, taskc->cpu, 0, &is_idle);
+		cpuc = lookup_cpu_ctx(cpu);
+		if (cpuc && taskc->dsq_index >= 0 && taskc->dsq_index < nr_dsqs_per_llc) {
+			dsq_id = cpu_dsq_id(taskc->dsq_index, cpuc);
+			scx_bpf_dsq_insert_vtime(p, dsq_id, slice_ns, vtime, enq_flags);
+			if (is_idle) {
+				stat_inc(P2DQ_STAT_IDLE);
+				scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+			}
+
+			ret->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
+			return;
+		}
+	}
+
 	dsq_id = cpu_dsq_id(taskc->dsq_index, cpuc);
-	scx_bpf_dsq_insert_vtime(p, dsq_id, slice_ns, vtime, enq_flags);
-	if (taskc->dsq_index < 0 || taskc->dsq_index >= nr_dsqs_per_llc) {
-		scx_bpf_error("can't happen");
-		return;
-	}
 
-	if (vtime > llcx->dsq_max_vtime[taskc->dsq_index]) {
-		llcx->dsq_max_vtime[taskc->dsq_index] = vtime;
-		trace("LLC[%d]DSQ[%d] max_vtime %llu", llcx->id, dsq_id, vtime);
-	}
-
-
-	llc_mask = cast_mask(llcx->cpumask);
-	if (!llc_mask) {
-		scx_bpf_error("invalid llc cpumask");
-		return;
-
-	}
-
-	idle_cpumask = scx_bpf_get_idle_cpumask();
-	cpu = bpf_cpumask_any_and_distribute(llc_mask, idle_cpumask);
-	scx_bpf_put_cpumask(idle_cpumask);
-
-	// If there is an idle CPU in the LLC kick it.
-	if (cpu < nr_cpus)
-		scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+	ret->kind = P2DQ_ENQUEUE_PROMISE_VTIME;
+	ret->vtime.dsq_id = dsq_id;
+	ret->vtime.enq_flags = enq_flags;
+	ret->vtime.slice_ns = slice_ns;
+	ret->vtime.vtime = vtime;
 }
 
+static __always_inline void complete_p2dq_enqueue(struct enqueue_promise *pro,
+						  struct task_struct *p)
+{
+	switch (pro->kind) {
+	case P2DQ_ENQUEUE_PROMISE_COMPLETE:
+		goto out;
+	case P2DQ_ENQUEUE_PROMISE_FIFO:
+		scx_bpf_dsq_insert(p, pro->fifo.dsq_id, pro->fifo.slice_ns,
+				   pro->fifo.enq_flags);
+		goto out;
+	case P2DQ_ENQUEUE_PROMISE_VTIME:
+		scx_bpf_dsq_insert_vtime(p, pro->vtime.dsq_id, pro->vtime.slice_ns,
+				         pro->vtime.vtime, pro->vtime.enq_flags);
+		goto out;
+	}
+out:
+	pro->kind = P2DQ_ENQUEUE_PROMISE_COMPLETE;
+}
 
-void BPF_STRUCT_OPS(p2dq_runnable, struct task_struct *p, u64 enq_flags)
+static __always_inline void p2dq_runnable_impl(struct task_struct *p, u64 enq_flags)
 {
 	struct task_ctx *wakee_ctx;
 
@@ -665,10 +800,13 @@ void BPF_STRUCT_OPS(p2dq_running, struct task_struct *p)
 		return;
 
 	if (taskc->llc_id != cpuc->llc_id) {
+		taskc->llc_runs = 0;
 		stat_inc(P2DQ_STAT_LLC_MIGRATION);
 		trace("RUNNING %d cpu %d->%d llc %d->%d",
 		      p->pid, cpuc->id, taskc->cpu,
 		      taskc->llc_id, llcx->id);
+	} else {
+		taskc->llc_runs += 1;
 	}
 	if (taskc->node_id != cpuc->node_id) {
 		stat_inc(P2DQ_STAT_NODE_MIGRATION);
@@ -686,8 +824,10 @@ void BPF_STRUCT_OPS(p2dq_running, struct task_struct *p)
 	else
 		p->scx.dsq_vtime = llcx->vtime;
 
-	if (sched_mode == MODE_PERFORMANCE && taskc->dsq_index == 0) {
-		scx_bpf_cpuperf_set(task_cpu, 1024);
+	// If the task is running in the least interactive DSQ, bump the
+	// frequency.
+	if (taskc->dsq_index == nr_dsqs_per_llc-1) {
+		scx_bpf_cpuperf_set(task_cpu, SCX_CPUPERF_ONE);
 	}
 }
 
@@ -697,7 +837,7 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 	struct task_ctx *taskc;
 	struct cpu_ctx *cpuc;
 	struct llc_ctx *llcx;
-	u64 used, last_dsq_slice_ns;
+	u64 used, scaled_used, last_dsq_slice_ns;
 	u64 now = scx_bpf_now();
 
 	if (!(taskc = lookup_task_ctx(p)))
@@ -717,10 +857,12 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 	}
 
 	used = now - taskc->last_run_at;
+	scaled_used = scale_by_task_weight_inverse(p, used);
+	p->scx.dsq_vtime += scaled_used;
 	taskc->last_dsq_id = taskc->dsq_id;
 	taskc->last_dsq_index = dsq_index;
-	__sync_fetch_and_add(&llcx->vtime, used);
-	__sync_fetch_and_add(&llcx->dsq_max_vtime[dsq_index], used);
+	__sync_fetch_and_add(&llcx->vtime, scaled_used);
+	__sync_fetch_and_add(&llcx->dsq_max_vtime[dsq_index], scaled_used);
 	__sync_fetch_and_add(&llcx->dsq_load[dsq_index], used);
 
 	// On stopping determine if the task can move to a longer DSQ by
@@ -749,9 +891,95 @@ void BPF_STRUCT_OPS(p2dq_stopping, struct task_struct *p, bool runnable)
 	}
 }
 
+static int __always_inline dispatch_cpu(u64 dsq_id, s32 cpu, struct llc_ctx *llcx, int dsq_index)
+{
+	struct task_struct *p;
+	int dispatched = 0;
+
+	if ((max_dsq_pick2 && dsq_index > 1) ||
+	    (min_nr_queued_pick2 > 0 &&
+	    scx_bpf_dsq_nr_queued(dsq_id) < min_nr_queued_pick2))
+		return -EINVAL;
+
+	bpf_for_each(scx_dsq, p, dsq_id, 0) {
+		/*
+		 * This is a workaround for the BPF verifier's pointer
+		 * validation limitations. Once the verifier gets smarter
+		 * we can remove this bpf_task_from_pid().
+		 */
+		p = bpf_task_from_pid(p->pid);
+		if (!p)
+			continue;
+
+		if (!bpf_cpumask_test_cpu(cpu, p->cpus_ptr)) {
+			bpf_task_release(p);
+			continue;
+		}
+
+		if (!__COMPAT_scx_bpf_dsq_move(BPF_FOR_EACH_ITER, p, SCX_DSQ_LOCAL_ON|cpu, SCX_ENQ_PREEMPT)) {
+			bpf_task_release(p);
+			continue;
+		}
+		dispatched += 1;
+		stat_inc(P2DQ_STAT_DISPATCH_PICK2);
+		bpf_task_release(p);
+		break;
+	}
+
+	return dispatched;
+}
 
 
-void BPF_STRUCT_OPS(p2dq_dispatch, s32 cpu, struct task_struct *prev)
+static __always_inline int dispatch_pick_two(s32 cpu, struct llc_ctx *cur_llcx, struct cpu_ctx *cpuc)
+{
+	struct llc_ctx *llcx, *left, *right;
+	u64 dsq_id;
+	int i;
+
+	// If on a single LLC there isn't anything left to try.
+	if (nr_llcs == 1 || dispatch_pick2_disable)
+		return -EINVAL;
+
+	// Special case when two llcs are present
+	if (nr_llcs == 2) {
+		left = lookup_llc_ctx(0);
+		right = lookup_llc_ctx(1);
+	} else {
+		left = rand_llc_ctx();
+		right = rand_llc_ctx();
+	}
+
+	// Last ditch effort try consuming from the most loaded DSQ.
+	llcx = pick_two_llc_ctx(cur_llcx, left, right);
+	if (!llcx)
+		return -EINVAL;
+
+	// The compat macro doesn't work properly, so on older kernels best
+	// effort by moving to local directly instead of iterating.
+	if (!bpf_ksym_exists(scx_bpf_dsq_move)) {
+		// Start with least interactive DSQs to avoid migrating
+		// interactive tasks.
+		bpf_for(i, 1, nr_dsqs_per_llc) {
+			if (scx_bpf_dsq_move_to_local(llcx->dsqs[nr_dsqs_per_llc - i])) {
+				stat_inc(P2DQ_STAT_DISPATCH_PICK2);
+				return 0;
+			}
+		}
+		return 0;
+	}
+
+	// Start with least interactive DSQs to avoid migrating interactive
+	// tasks.
+	bpf_for(i, 1, nr_dsqs_per_llc) {
+		dsq_id = llcx->dsqs[nr_dsqs_per_llc - i];
+		if (dispatch_cpu(dsq_id, cpu, llcx, nr_dsqs_per_llc - i) > 0)
+			return 0;
+	}
+	return 0;
+}
+
+
+static __always_inline void p2dq_dispatch_impl(s32 cpu, struct task_struct *prev)
 {
 	struct cpu_ctx *cpuc;
 	struct llc_ctx *llcx;
@@ -793,21 +1021,7 @@ void BPF_STRUCT_OPS(p2dq_dispatch, s32 cpu, struct task_struct *prev)
 		    return;
 	}
 
-	// If on a single LLC there isn't anything left to try.
-	if (nr_llcs == 1)
-		return;
-
-	// Last ditch effort try consuming from the most loaded DSQ.
-	llcx = pick_two_llc_ctx(rand_llc_ctx(), rand_llc_ctx(), false);
-	if (!llcx || cpuc->dsq_index < 0 || cpuc->dsq_index > nr_dsqs_per_llc)
-		return;
-
-	// Start with least interactive DSQs to avoid migrating interactive
-	// tasks.
-	bpf_for(i, 1, nr_dsqs_per_llc) {
-		if (scx_bpf_dsq_move_to_local(llcx->dsqs[nr_dsqs_per_llc - i]))
-		    return;
-	}
+	dispatch_pick_two(cpu, llcx, cpuc);
 }
 
 void BPF_STRUCT_OPS(p2dq_set_cpumask, struct task_struct *p,
@@ -818,7 +1032,7 @@ void BPF_STRUCT_OPS(p2dq_set_cpumask, struct task_struct *p,
 	if (!(taskc = lookup_task_ctx(p)) || !all_cpumask)
 		return;
 
-	taskc->all_cpus = p->nr_cpus_allowed == nr_cpus;
+	taskc->all_cpus = p->cpus_ptr == &p->cpus_mask && p->nr_cpus_allowed == nr_cpus;
 }
 
 s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
@@ -838,6 +1052,10 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 		return -ENOMEM;
 	}
 
+	if (!(cpuc = lookup_cpu_ctx(task_cpu)) ||
+	    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
+		return -EINVAL;
+
 	if (!(cpumask = bpf_cpumask_create())) {
 		scx_bpf_error("task_ctx allocation failure");
 		return -ENOMEM;
@@ -849,10 +1067,6 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 		return -EINVAL;
 	}
 
-	if (!(cpuc = lookup_cpu_ctx(task_cpu)) ||
-	    !(llcx = lookup_llc_ctx(cpuc->llc_id)))
-		return -EINVAL;
-
 	taskc->cpu = task_cpu;
 	taskc->dsq_id = SCX_DSQ_INVALID;
 	taskc->llc_id = cpuc->llc_id;
@@ -860,7 +1074,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 	taskc->dsq_index = init_dsq_index;
 	taskc->last_dsq_index = init_dsq_index;
 	taskc->runnable = true;
-	taskc->all_cpus = p->nr_cpus_allowed == nr_cpus;
+	taskc->all_cpus = p->cpus_ptr == &p->cpus_mask && p->nr_cpus_allowed == nr_cpus;
 	p->scx.dsq_vtime = llcx->vtime;
 
 	return 0;
@@ -869,7 +1083,7 @@ s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init_task, struct task_struct *p,
 
 static int init_llc(u32 llc_id)
 {
-	struct bpf_cpumask *cpumask, *big_cpumask;
+	struct bpf_cpumask *cpumask, *big_cpumask, *tmp_cpumask;
 	struct llc_ctx *llcx;
 
 	llcx = bpf_map_lookup_elem(&llc_ctxs, &llc_id);
@@ -911,6 +1125,18 @@ static int init_llc(u32 llc_id)
 		bpf_cpumask_release(big_cpumask);
 	}
 
+	tmp_cpumask = bpf_cpumask_create();
+	if (!tmp_cpumask) {
+		scx_bpf_error("failed to create tmp cpumask");
+		return -ENOMEM;
+	}
+
+	tmp_cpumask = bpf_kptr_xchg(&llcx->tmp_cpumask, tmp_cpumask);
+	if (tmp_cpumask) {
+		scx_bpf_error("kptr already had cpumask");
+		bpf_cpumask_release(tmp_cpumask);
+	}
+
 	return 0;
 }
 
@@ -1162,7 +1388,7 @@ s32 static start_timers(void)
 	return 0;
 }
 
-s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init)
+static __always_inline s32 p2dq_init_impl()
 {
 	int i, ret;
 	struct bpf_cpumask *tmp_cpumask, *tmp_big_cpumask;
@@ -1260,6 +1486,34 @@ void BPF_STRUCT_OPS(p2dq_exit, struct scx_exit_info *ei)
 	UEI_RECORD(uei, ei);
 }
 
+#if P2DQ_CREATE_STRUCT_OPS
+void BPF_STRUCT_OPS(p2dq_runnable, struct task_struct *p, u64 enq_flags)
+{
+	return p2dq_runnable_impl(p, enq_flags);
+}
+
+s32 BPF_STRUCT_OPS_SLEEPABLE(p2dq_init)
+{
+	return p2dq_init_impl();
+}
+
+void BPF_STRUCT_OPS(p2dq_enqueue, struct task_struct *p __arg_trusted, u64 enq_flags)
+{
+	struct enqueue_promise pro;
+	async_p2dq_enqueue(&pro, p, enq_flags);
+	complete_p2dq_enqueue(&pro, p);
+}
+
+void BPF_STRUCT_OPS(p2dq_dispatch, s32 cpu, struct task_struct *prev)
+{
+	return p2dq_dispatch_impl(cpu, prev);
+}
+
+s32 BPF_STRUCT_OPS(p2dq_select_cpu, struct task_struct *p, s32 prev_cpu, u64 wake_flags)
+{
+	return p2dq_select_cpu_impl(p, prev_cpu, wake_flags);
+}
+
 SCX_OPS_DEFINE(p2dq,
 	       .select_cpu		= (void *)p2dq_select_cpu,
 	       .enqueue			= (void *)p2dq_enqueue,
@@ -1273,3 +1527,4 @@ SCX_OPS_DEFINE(p2dq,
 	       .exit			= (void *)p2dq_exit,
 	       .timeout_ms		= 20000,
 	       .name			= "p2dq");
+#endif
