@@ -456,7 +456,7 @@ static void dom_dcycle_xfer_task(struct task_struct *p, struct task_ctx *taskc,
 
 static u64 dom_min_vruntime(dom_ptr domc)
 {
-	return READ_ONCE_ARENA(u64, domc->min_vruntime);
+	return READ_ONCE(domc->min_vruntime);
 }
 
 int dom_xfer_task(struct task_struct *p __arg_trusted, u32 new_dom_id, u64 now)
@@ -1279,7 +1279,7 @@ static u64 node_dom_mask(u32 node_id)
 	return mask;
 }
 
-#ifndef __TARGET_ARCH_arm64
+#if !defined(__TARGET_ARCH_arm64) && ! defined(__TARGET_ARCH_riscv)
 /*
  * Sets the preferred domain mask according to the mempolicy. See man(2)
  * set_mempolicy for more details on mempolicy.
@@ -1458,7 +1458,7 @@ static void running_update_vtime(struct task_struct *p,
 
 	bpf_spin_lock(lock);
 	if (time_before(dom_min_vruntime(domc), p->scx.dsq_vtime))
-		WRITE_ONCE_ARENA(u64, domc->min_vruntime, p->scx.dsq_vtime);
+		WRITE_ONCE(domc->min_vruntime, p->scx.dsq_vtime);
 	bpf_spin_unlock(lock);
 }
 
