@@ -161,14 +161,6 @@ struct task_ctx *try_lookup_task_ctx(const struct task_struct *p)
  */
 static u64 vtime_now;
 
-/*
- * Scale value inversely proportional to the weight of @p.
- */
-static u64 scale_inverse_fair(const struct task_struct *p, u64 value)
-{
-	return value * 100 / p->scx.weight;
-}
-
 /*
  * Update and return the task's deadline.
  */
@@ -191,7 +183,7 @@ static u64 task_deadline(const struct task_struct *p, struct task_ctx *tctx)
 	/*
 	 * Add the execution vruntime to the deadline.
 	 */
-	return tctx->deadline + scale_inverse_fair(p, tctx->exec_runtime);
+	return tctx->deadline + scale_by_task_weight_inverse(p, tctx->exec_runtime);
 }
 
 /*
@@ -564,7 +556,7 @@ void BPF_STRUCT_OPS(tickless_stopping, struct task_struct *p, bool runnable)
 	/*
 	 * Update task's vruntime.
 	 */
-	tctx->deadline += scale_inverse_fair(p, slice);
+	tctx->deadline += scale_by_task_weight_inverse(p, slice);
 }
 
 /*
