@@ -25,8 +25,9 @@ use crate::APP;
 use crate::LICENSE;
 use crate::SCHED_NAME_PATH;
 use crate::{
-    Action, CpuhpAction, GpuMemAction, IPIAction, SchedCpuPerfSetAction, SchedSwitchAction,
-    SchedWakeupAction, SchedWakingAction, SoftIRQAction, TraceStartedAction, TraceStoppedAction,
+    Action, CpuhpEnterAction, CpuhpExitAction, ExecAction, ExitAction, ForkAction, GpuMemAction,
+    HwPressureAction, IPIAction, SchedCpuPerfSetAction, SchedSwitchAction, SchedWakeupAction,
+    SchedWakingAction, SoftIRQAction, TraceStartedAction, TraceStoppedAction,
 };
 
 use anyhow::Result;
@@ -50,12 +51,14 @@ use ratatui::{
 use regex::Regex;
 use scx_stats::prelude::StatsClient;
 use scx_utils::misc::read_file_usize;
+use scx_utils::scx_enums;
 use scx_utils::Topology;
 use serde_json::Value as JsonValue;
 use tokio::sync::mpsc::UnboundedSender;
 use tokio::sync::Mutex;
 
 use std::collections::BTreeMap;
+use std::os::fd::{AsFd, AsRawFd};
 use std::path::Path;
 use std::sync::atomic::AtomicBool;
 use std::sync::atomic::Ordering;
@@ -66,6 +69,7 @@ const DSQ_VTIME_CUTOFF: u64 = 1_000_000_000_000_000;
 /// App is the struct for scxtop application state.
 pub struct App<'a> {
     config: Config,
+    hw_pressure: bool,
     localize: bool,
     locale: SystemLocale,
     stats_client: Arc<Mutex<StatsClient>>,
@@ -190,9 +194,14 @@ impl<'a> App<'a> {
         let trace_file_prefix = config.trace_file_prefix().to_string();
         let trace_manager = PerfettoTraceManager::new(trace_file_prefix, None);
 
+        // There isn't a 'is_loaded' method on a prog in libbpf-rs so do the next best thing and
+        // try to infer from the fd
+        let hw_pressure = skel.progs.on_hw_pressure_update.as_fd().as_raw_fd() > 0;
+
         let app = Self {
             config,
             localize: true,
+            hw_pressure,
             locale: SystemLocale::default()?,
             stats_client: Arc::new(Mutex::new(stats_client)),
             sched_stats_raw: "".to_string(),
@@ -500,21 +509,23 @@ impl<'a> App<'a> {
 
     /// Generates a CPU bar chart.
     fn cpu_bar(&self, cpu: usize, event: &str) -> Bar {
-        let value = self
-            .cpu_data
-            .get(&cpu)
-            .unwrap()
+        let cpu_data = self.cpu_data.get(&cpu).unwrap();
+        let value = cpu_data
             .event_data_immut(event)
             .last()
             .copied()
             .unwrap_or(0_u64);
+        let hw_pressure = cpu_data
+            .event_data_immut("hw_pressure")
+            .last()
+            .copied()
+            .unwrap_or(0);
         Bar::default()
             .value(value)
             .label(Line::from(format!(
-                "{}{}",
+                "{}{}{}",
                 cpu,
                 if self.collect_cpu_freq {
-                    let cpu_data = self.cpu_data.get(&cpu).unwrap();
                     format!(
                         " {}",
                         format_hz(
@@ -527,6 +538,11 @@ impl<'a> App<'a> {
                     )
                 } else {
                     "".to_string()
+                },
+                if self.hw_pressure && hw_pressure > 0 {
+                    format!("{}", hw_pressure)
+                } else {
+                    "".to_string()
                 }
             )))
             .text_value(if self.localize {
@@ -540,6 +556,7 @@ impl<'a> App<'a> {
     fn cpu_sparkline(&self, cpu: usize, max: u64, borders: Borders, small: bool) -> Sparkline {
         let mut perf: u64 = 0;
         let mut cpu_freq: u64 = 0;
+        let mut hw_pressure: u64 = 0;
         let data = if self.cpu_data.contains_key(&cpu) {
             let cpu_data = self.cpu_data.get(&cpu).unwrap();
             perf = cpu_data
@@ -554,6 +571,11 @@ impl<'a> App<'a> {
                     .copied()
                     .unwrap_or(0);
             }
+            hw_pressure = cpu_data
+                .event_data_immut("hw_pressure")
+                .last()
+                .copied()
+                .unwrap_or(0);
             cpu_data.event_data_immut(self.active_event.event_name())
         } else {
             Vec::new()
@@ -567,7 +589,7 @@ impl<'a> App<'a> {
             .block(
                 Block::new()
                     .title(format!(
-                        "{} perf({}){}",
+                        "{} perf({}){}{}",
                         cpu,
                         if perf == 0 {
                             "".to_string()
@@ -578,6 +600,11 @@ impl<'a> App<'a> {
                             format!(" {}", format_hz(cpu_freq))
                         } else {
                             "".to_string()
+                        },
+                        if self.hw_pressure && hw_pressure > 0 {
+                            format!(" hw_pressure({})", hw_pressure)
+                        } else {
+                            "".to_string()
                         }
                     ))
                     .borders(borders)
@@ -2128,6 +2155,9 @@ impl<'a> App<'a> {
             self.skel.progs.on_softirq_entry.attach()?,
             self.skel.progs.on_softirq_exit.attach()?,
             self.skel.progs.on_ipi_send_cpu.attach()?,
+            self.skel.progs.on_sched_fork.attach()?,
+            self.skel.progs.on_sched_exec.attach()?,
+            self.skel.progs.on_sched_exit.attach()?,
         ];
 
         Ok(())
@@ -2254,6 +2284,24 @@ impl<'a> App<'a> {
         cpu_data.add_event_data("perf", perf as u64);
     }
 
+    fn on_exec(&mut self, action: &ExecAction) {
+        if self.state == AppState::Tracing && action.ts > self.trace_start {
+            self.trace_manager.on_exec(action);
+        }
+    }
+
+    fn on_exit(&mut self, action: &ExitAction) {
+        if self.state == AppState::Tracing && action.ts > self.trace_start {
+            self.trace_manager.on_exit(action);
+        }
+    }
+
+    fn on_fork(&mut self, action: &ForkAction) {
+        if self.state == AppState::Tracing && action.ts > self.trace_start {
+            self.trace_manager.on_fork(action);
+        }
+    }
+
     /// Updates the app when a task wakes.
     fn on_sched_wakeup(&mut self, action: &SchedWakeupAction) {
         if self.state == AppState::Tracing && action.ts > self.trace_start {
@@ -2296,32 +2344,37 @@ impl<'a> App<'a> {
             0,
             self.max_cpu_events,
         ));
-        cpu_data.add_event_data("dsq_lat_us", *next_dsq_lat_us);
-        let next_dsq_data = self
-            .dsq_data
-            .entry(*next_dsq_id)
-            .or_insert(EventData::new(self.max_cpu_events));
-        next_dsq_data.add_event_data("dsq_lat_us", *next_dsq_lat_us);
-        if *next_dsq_vtime > 0 {
-            // vtime is special because we want the delta
-            let last = next_dsq_data
-                .event_data_immut("dsq_vtime_delta")
-                .last()
-                .copied()
-                .unwrap_or(0_u64);
-            if next_dsq_vtime - last < DSQ_VTIME_CUTOFF {
-                next_dsq_data.add_event_data(
-                    "dsq_vtime_delta",
-                    if last > 0 { *next_dsq_vtime - last } else { 0 },
-                );
+
+        if *next_dsq_id != scx_enums.SCX_DSQ_INVALID && *next_dsq_lat_us > 0 {
+            cpu_data.add_event_data("dsq_lat_us", *next_dsq_lat_us);
+            let next_dsq_data = self
+                .dsq_data
+                .entry(*next_dsq_id)
+                .or_insert(EventData::new(self.max_cpu_events));
+            next_dsq_data.add_event_data("dsq_lat_us", *next_dsq_lat_us);
+            if *next_dsq_vtime > 0 {
+                // vtime is special because we want the delta
+                let last = next_dsq_data
+                    .event_data_immut("dsq_vtime_delta")
+                    .last()
+                    .copied()
+                    .unwrap_or(0_u64);
+                if next_dsq_vtime - last < DSQ_VTIME_CUTOFF {
+                    next_dsq_data.add_event_data(
+                        "dsq_vtime_delta",
+                        if last > 0 { *next_dsq_vtime - last } else { 0 },
+                    );
+                }
             }
         }
 
-        let prev_dsq_data = self
-            .dsq_data
-            .entry(*prev_dsq_id)
-            .or_insert(EventData::new(self.max_cpu_events));
-        prev_dsq_data.add_event_data("dsq_slice_consumed", *prev_used_slice_ns);
+        if *prev_dsq_id != scx_enums.SCX_DSQ_INVALID && *prev_used_slice_ns > 0 {
+            let prev_dsq_data = self
+                .dsq_data
+                .entry(*prev_dsq_id)
+                .or_insert(EventData::new(self.max_cpu_events));
+            prev_dsq_data.add_event_data("dsq_slice_consumed", *prev_used_slice_ns);
+        }
     }
 
     /// Handles softirq events.
@@ -2345,12 +2398,33 @@ impl<'a> App<'a> {
     }
 
     /// Handles cpu hotplug events.
-    pub fn on_cpu_hp(&mut self, action: &CpuhpAction) {
+    pub fn on_cpu_hp_enter(&mut self, action: &CpuhpEnterAction) {
         if self.state == AppState::Tracing && action.ts > self.trace_start {
-            self.trace_manager.on_cpu_hp(action);
+            self.trace_manager.on_cpu_hp_enter(action);
         }
     }
 
+    pub fn on_cpu_hp_exit(&mut self, action: &CpuhpExitAction) {
+        if self.state == AppState::Tracing && action.ts > self.trace_start {
+            self.trace_manager.on_cpu_hp_exit(action);
+        }
+    }
+
+    /// Handles hardware pressure events.
+    pub fn on_hw_pressure(&mut self, action: &HwPressureAction) {
+        let HwPressureAction { cpu, hw_pressure } = action;
+
+        let cpu_data = self.cpu_data.entry(*cpu as usize).or_insert(CpuData::new(
+            *cpu as usize,
+            0,
+            0,
+            0,
+            self.max_cpu_events,
+        ));
+
+        cpu_data.add_event_data("hw_pressure", *hw_pressure);
+    }
+
     /// Updates the bpf bpf sampling rate.
     pub fn update_bpf_sample_rate(&mut self, sample_rate: u32) {
         self.skel.maps.data_data.sample_rate = sample_rate;
@@ -2429,14 +2503,29 @@ impl<'a> App<'a> {
             Action::SoftIRQ(a) => {
                 self.on_softirq(a);
             }
+            Action::Exec(a) => {
+                self.on_exec(a);
+            }
+            Action::Exit(a) => {
+                self.on_exit(a);
+            }
+            Action::Fork(a) => {
+                self.on_fork(a);
+            }
             Action::IPI(a) => {
                 self.on_ipi(a);
             }
             Action::GpuMem(a) => {
                 self.on_gpu_mem(a);
             }
-            Action::Cpuhp(a) => {
-                self.on_cpu_hp(a);
+            Action::CpuhpEnter(a) => {
+                self.on_cpu_hp_enter(a);
+            }
+            Action::CpuhpExit(a) => {
+                self.on_cpu_hp_exit(a);
+            }
+            Action::HwPressure(a) => {
+                self.on_hw_pressure(a);
             }
             Action::ClearEvent => self.stop_perf_events(),
             Action::ChangeTheme => {
@@ -2449,6 +2538,7 @@ impl<'a> App<'a> {
             Action::ToggleCpuFreq => self.collect_cpu_freq = !self.collect_cpu_freq,
             Action::ToggleUncoreFreq => self.collect_uncore_freq = !self.collect_uncore_freq,
             Action::ToggleLocalization => self.localize = !self.localize,
+            Action::ToggleHwPressure => self.hw_pressure = !self.hw_pressure,
             Action::IncBpfSampleRate => {
                 let sample_rate = self.skel.maps.data_data.sample_rate;
                 if sample_rate == 0 {
