@@ -242,13 +242,13 @@ int BPF_KPROBE(on_sched_cpu_perf, s32 cpu, u32 perf)
 	return 0;
 }
 
-SEC("?kprobe/scx_bpf_dsq_insert_vtime")
+SEC("kprobe/scx_bpf_dsq_insert_vtime")
 int BPF_KPROBE(scx_insert_vtime, struct task_struct *p, u64 dsq, u64 slice_ns, u64 vtime)
 {
 	return update_task_ctx(p, dsq, vtime, slice_ns);
 }
 
-SEC("?kprobe/scx_bpf_dispatch_vtime")
+SEC("kprobe/scx_bpf_dispatch_vtime")
 int BPF_KPROBE(scx_dispatch_vtime, struct task_struct *p, u64 dsq, u64 slice_ns, u64 vtime)
 {
 	return update_task_ctx(p, dsq, vtime, slice_ns);
@@ -272,13 +272,13 @@ static int on_insert(struct task_struct *p, u64 dsq)
 }
 
 
-SEC("?kprobe/scx_bpf_dsq_insert")
+SEC("kprobe/scx_bpf_dsq_insert")
 int BPF_KPROBE(scx_insert, struct task_struct *p, u64 dsq)
 {
 	return on_insert(p, dsq);
 }
 
-SEC("?kprobe/scx_bpf_dispatch")
+SEC("kprobe/scx_bpf_dispatch")
 int BPF_KPROBE(scx_dispatch, struct task_struct *p, u64 dsq)
 {
 	return on_insert(p, dsq);
@@ -294,20 +294,21 @@ static int on_dsq_move(struct task_struct *p, u64 dsq)
 	if (!(tctx = try_lookup_task_ctx(p)))
 		return -ENOENT;
 
+	tctx->dsq_insert_time = bpf_ktime_get_ns();
 	tctx->dsq_id = dsq;
 	tctx->dsq_vtime = 0;
 
 	return 0;
 }
 
-SEC("?kprobe/scx_bpf_dsq_move")
+SEC("kprobe/scx_bpf_dsq_move")
 int BPF_KPROBE(scx_dsq_move, struct bpf_iter_scx_dsq *it__iter,
 	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move(p, dsq_id);
 }
 
-SEC("?kprobe/scx_bpf_dispatch_from_dsq")
+SEC("kprobe/scx_bpf_dispatch_from_dsq")
 int BPF_KPROBE(scx_dispatch_from_dsq, struct bpf_iter_scx_dsq *it__iter,
 	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
@@ -324,20 +325,21 @@ static int on_dsq_move_vtime(struct task_struct *p, u64 dsq)
 	if (!(tctx = try_lookup_task_ctx(p)))
 		return -ENOENT;
 
+	tctx->dsq_insert_time = bpf_ktime_get_ns();
 	tctx->dsq_id = dsq;
 	bpf_core_read(&tctx->dsq_vtime, sizeof(u64), &p->scx.dsq_vtime);
 
 	return 0;
 }
 
-SEC("?kprobe/scx_bpf_dsq_move_vtime")
+SEC("kprobe/scx_bpf_dsq_move_vtime")
 int BPF_KPROBE(scx_dsq_move_vtime, struct bpf_iter_scx_dsq *it__iter,
 	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
 	return on_dsq_move_vtime(p, dsq_id);
 }
 
-SEC("?kprobe/scx_bpf_dispatch_vtime_from_dsq")
+SEC("kprobe/scx_bpf_dispatch_vtime_from_dsq")
 int BPF_KPROBE(scx_dispatch_vtime_from_dsq, struct bpf_iter_scx_dsq *it__iter,
 	       struct task_struct *p, u64 dsq_id, u64 enq_flags)
 {
@@ -359,14 +361,14 @@ static int on_move_set_slice(struct task_struct *p, u64 slice)
 	return 0;
 }
 
-SEC("?kprobe/scx_bpf_dsq_move_set_slice")
+SEC("kprobe/scx_bpf_dsq_move_set_slice")
 int BPF_KPROBE(scx_dsq_move_set_slice, struct bpf_iter_scx_dsq *it__iter, u64 slice)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_slice(NULL, slice);
 }
 
-SEC("?kprobe/scx_bpf_dispatch_from_dsq_set_slice")
+SEC("kprobe/scx_bpf_dispatch_from_dsq_set_slice")
 int BPF_KPROBE(scx_dispatch_from_dsq_set_slice, struct bpf_iter_scx_dsq *it__iter,
 	       u64 slice)
 {
@@ -389,14 +391,14 @@ static int on_move_set_vtime(struct task_struct *p, u64 vtime)
 	return 0;
 }
 
-SEC("?kprobe/scx_bpf_dsq_move_set_vtime")
+SEC("kprobe/scx_bpf_dsq_move_set_vtime")
 int BPF_KPROBE(scx_dsq_move_set_vtime, struct bpf_iter_scx_dsq *it__iter, u64 vtime)
 {
 	// TODO: figure out how to return task from iterator without consuming.
 	return on_move_set_vtime(NULL, vtime);
 }
 
-SEC("?kprobe/scx_bpf_dispatch_from_dsq_set_vtime")
+SEC("kprobe/scx_bpf_dispatch_from_dsq_set_vtime")
 int BPF_KPROBE(scx_dispatch_from_dsq_set_vtime, struct bpf_iter_scx_dsq *it__iter, u64 vtime)
 {
 	// TODO: figure out how to return task from iterator without consuming.
@@ -411,16 +413,18 @@ static __always_inline int __on_sched_wakeup(struct task_struct *p)
 	if (!p || !should_sample())
 		return 0;
 
+	u64 now = bpf_ktime_get_ns();
 	tctx = try_lookup_task_ctx(p);
-	if (!tctx)
-		return 0;
 
 	if (!(event = try_reserve_event()))
 		return 0;
 
-	tctx->wakeup_ts = bpf_ktime_get_ns();
 	event->type = SCHED_WAKEUP;
-	event->ts = tctx->wakeup_ts;
+
+	if (tctx)
+		tctx->wakeup_ts = now;
+
+	event->ts = now;
 	event->cpu = bpf_get_smp_processor_id();
 	event->event.wakeup.pid = p->pid;
 	event->event.wakeup.prio = (int)p->prio;
@@ -498,9 +502,9 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 		event->event.sched_switch.next_pid = next->pid;
 		event->event.sched_switch.next_tgid = next->tgid;
 		event->event.sched_switch.next_prio = (int)next->prio;
-		if (next_tctx) {
-			event->event.sched_switch.next_dsq_id = next_tctx->dsq_id;
+		if (next_tctx && next_tctx->dsq_insert_time > 0) {
 			event->event.sched_switch.next_dsq_lat_us = (now - next_tctx->dsq_insert_time) / 1000;
+			event->event.sched_switch.next_dsq_id = next_tctx->dsq_id;
 			event->event.sched_switch.next_dsq_nr = scx_bpf_dsq_nr_queued(next_tctx->dsq_id);
 			/*
 			 * XXX: if a task gets moved to another dsq and the vtime is updated
@@ -527,7 +531,7 @@ int BPF_PROG(on_sched_switch, bool preempt, struct task_struct *prev,
 		event->event.sched_switch.prev_tgid = prev->tgid;
 		event->event.sched_switch.prev_prio = (int)prev->prio;
 		event->event.sched_switch.prev_state = prev_state;
-		if (prev_tctx) {
+		if (prev_tctx && prev_tctx->last_run_ns > 0) {
 			event->event.sched_switch.prev_used_slice_ns = prev_tctx->last_run_ns - now;
 			event->event.sched_switch.prev_dsq_id = prev_tctx->dsq_id;
 			event->event.sched_switch.prev_slice_ns = prev_tctx->slice_ns;
@@ -778,6 +782,76 @@ int BPF_PROG(on_ipi_send_cpu, u32 cpu, void *callsite, void *callback)
 	return 0;
 }
 
+SEC("tp_btf/sched_process_exit")
+int BPF_PROG(on_sched_exit, struct task_struct *task)
+{
+	struct bpf_event *event;
+
+	if (!enable_bpf_events || !should_sample())
+		return 0;
+
+	if (!(event = try_reserve_event()))
+		return -ENOMEM;
+
+	event->type = EXIT;
+	event->cpu = bpf_get_smp_processor_id();
+	event->ts = bpf_ktime_get_ns();
+	event->event.exit.pid = BPF_CORE_READ(task, pid);
+	event->event.exit.tgid = BPF_CORE_READ(task, tgid);
+	event->event.exit.prio = BPF_CORE_READ(task, prio);
+	__builtin_memcpy_inline(&event->event.exit.comm, &task->comm, MAX_COMM);
+
+	bpf_ringbuf_submit(event, 0);
+
+	return 0;
+}
+
+SEC("tp_btf/sched_process_fork")
+int BPF_PROG(on_sched_fork, struct task_struct *parent, struct task_struct *child)
+{
+	struct bpf_event *event;
+
+	if (!enable_bpf_events || !should_sample())
+		return 0;
+
+	if (!(event = try_reserve_event()))
+		return -ENOMEM;
+
+	event->type = FORK;
+	event->cpu = bpf_get_smp_processor_id();
+	event->ts = bpf_ktime_get_ns();
+	event->event.fork.parent_pid = BPF_CORE_READ(parent, pid);
+	event->event.fork.child_pid = BPF_CORE_READ(child, pid);
+	__builtin_memcpy_inline(&event->event.fork.parent_comm, &parent->comm, MAX_COMM);
+	__builtin_memcpy_inline(&event->event.fork.child_comm, &child->comm, MAX_COMM);
+
+	bpf_ringbuf_submit(event, 0);
+
+	return 0;
+}
+
+SEC("tp_btf/sched_process_exec")
+int BPF_PROG(on_sched_exec, struct task_struct *p, u32 old_pid, struct linux_binprm *prm)
+{
+	struct bpf_event *event;
+
+	if (!enable_bpf_events || !should_sample())
+		return 0;
+
+	if (!(event = try_reserve_event()))
+		return -ENOMEM;
+
+	event->type = EXEC;
+	event->cpu = bpf_get_smp_processor_id();
+	event->ts = bpf_ktime_get_ns();
+	event->event.exec.old_pid = old_pid;
+	event->event.exec.pid = BPF_CORE_READ(p, pid);
+
+	bpf_ringbuf_submit(event, 0);
+
+	return 0;
+}
+
 SEC("?tp_btf/gpu_mem_total")
 int BPF_PROG(on_gpu_memory_total, u32 gpu, u32 pid, u64 size)
 {
@@ -813,7 +887,7 @@ int BPF_PROG(on_cpuhp_enter, u32 cpu, int target, int state)
 	if (!(event = try_reserve_event()))
 		return -ENOMEM;
 
-	event->type = CPU_HP;
+	event->type = CPU_HP_ENTER;
 	event->cpu = bpf_get_smp_processor_id();
 	event->ts = bpf_ktime_get_ns();
 	event->event.chp.cpu = cpu;
@@ -830,6 +904,58 @@ int BPF_PROG(on_cpuhp_enter, u32 cpu, int target, int state)
 	return 0;
 }
 
+SEC("tp_btf/cpuhp_exit")
+int BPF_PROG(on_cpuhp_exit, u32 cpu, int state, int idx, int ret)
+{
+	struct bpf_event *event;
+	struct task_struct *p;
+
+	if (!enable_bpf_events || !should_sample())
+		return 0;
+
+	if (!(event = try_reserve_event()))
+		return -ENOMEM;
+
+	event->type = CPU_HP_EXIT;
+	event->cpu = bpf_get_smp_processor_id();
+	event->ts = bpf_ktime_get_ns();
+	event->event.cxp.cpu = cpu;
+	event->event.cxp.state = state;
+	event->event.cxp.state = idx;
+	event->event.cxp.state = ret;
+	p = (struct task_struct *)bpf_get_current_task();
+	if (p)
+		event->event.cxp.pid = BPF_CORE_READ(p, pid);
+	else
+		event->event.cxp.pid = 0;
+
+	bpf_ringbuf_submit(event, 0);
+
+	return 0;
+}
+
+SEC("?tp_btf/hw_pressure_update")
+int BPF_PROG(on_hw_pressure_update, u32 cpu, u64 hw_pressure)
+{
+	struct bpf_event *event;
+
+	if (!enable_bpf_events || !should_sample())
+		return 0;
+
+	if (!(event = try_reserve_event()))
+		return -ENOMEM;
+
+	event->type = HW_PRESSURE;
+	event->cpu = bpf_get_smp_processor_id();
+	event->ts = bpf_ktime_get_ns();
+	event->event.hwp.hw_pressure = hw_pressure;
+	event->event.hwp.cpu = cpu;
+
+	bpf_ringbuf_submit(event, 0);
+
+	return 0;
+}
+
 SEC("syscall")
 int BPF_PROG(scxtop_init)
 {
